"""
The View classes render the data returned by a Pipeline as a Panel
object.
"""
from __future__ import annotations

import sys

from io import BytesIO, StringIO
from typing import (
    IO, TYPE_CHECKING, Any, ClassVar, Dict, List,
)
from weakref import WeakKeyDictionary

import numpy as np
import panel as pn
import param  # type: ignore

from bokeh.models import NumeralTickFormatter  # type: ignore
from panel.pane.base import PaneBase
from panel.pane.perspective import (
    THEMES as _PERSPECTIVE_THEMES, Plugin as _PerspectivePlugin,
)
from panel.param import Param
from panel.viewable import Viewable, Viewer

from ..base import MultiTypeComponent
from ..config import _INDICATORS
from ..downloads import Download
from ..filters.base import Filter, ParamFilter
from ..pipeline import Pipeline
from ..state import state
from ..transforms.base import Transform
from ..transforms.sql import SQLTransform
from ..util import (
    VARIABLE_RE, catch_and_notify, immediate_dispatch, is_ref,
    resolve_module_reference,
)
from ..validation import ValidationError

if TYPE_CHECKING:
    from bokeh.document import Document  # type: ignore
    from holoviews.selection import link_selections  # type: ignore

DOWNLOAD_FORMATS = ['csv', 'xlsx', 'json', 'parquet']


class View(MultiTypeComponent, Viewer):
    """
    `View` components provide a visual representation for the data returned
    by a :class:`lumen.source.base.Source` or :class:`lumen.pipeline.Pipeline`.

    The `View` must return a Panel object or an object that can be
    rendered by Panel. The base class provides methods which query the
    the provided :class:`lumen.pipeline.Pipeline`.

    Subclasses should use these methods to query the data and return
    a Viewable Panel object in the `get_panel` method.
    """

    controls = param.List(default=[], doc="""
        Parameters that should be exposed as widgets in the UI.""")

    download = param.ClassSelector(class_=Download, default=Download(), doc="""
        The download objects determines whether and how the source tables
        can be downloaded.""")

    loading_indicator = param.Boolean(default=True, constant=True, doc="""
        Whether to display a loading indicator on the View when the
        Pipeline is refreshing the data.""")

    limit = param.Integer(default=None, doc="""
        Limits the number of rows that are rendered.""")

    pipeline = param.ClassSelector(class_=Pipeline, doc="""
        The data pipeline that drives the View.""")

    rerender = param.Event(default=False, doc="""
        An event that is triggered whenever the View requests a re-render.""")

    selection_group = param.String(default=None, doc="""
        Declares a selection group the plot is part of. This feature
        requires the separate HoloViews library.""")

    title = param.String(default=None, doc="""
        The title of the view.""")

    field = param.Selector(doc="The field being visualized.")

    view_type: ClassVar[str | None] = None

    # Panel extension to load to render this View
    _extension: ClassVar[str | None] = None

    # Parameters which reference fields in the table
    _field_params: ClassVar[List[str]] = ['field']

    _requires_source: ClassVar[bool] = True

    _selections: ClassVar[WeakKeyDictionary[Document, link_selections]] = WeakKeyDictionary()

    _supports_selections: ClassVar[bool] = False

    __abstract = True

    def __init__(self, **params):
        self._cache = None
        self._ls = None
        self._panel = None
        self._updates = None
        refs = params.pop('refs', {})
        self.kwargs = {k: v for k, v in params.items() if k not in self.param}

        # Populate field selector parameters
        params = {k: v for k, v in params.items() if k in self.param}
        pipeline = params.pop('pipeline', None)
        if pipeline is None:
            raise ValueError("Views must declare a Pipeline.")
        if isinstance(params.get("download"), str):
            *filenames, ext = params.get("download").split(".")
            filename = ".".join(filenames) or None
            params["download"] = Download(filename=filename, format=ext)
        fields = list(pipeline.schema)
        for fp in self._field_params:
            if isinstance(self.param[fp], param.Selector):
                self.param[fp].objects = fields
        pipeline.param.watch(self.update, 'data')
        if self.loading_indicator:
            pipeline._update_widget.param.watch(self._update_loading, 'loading')
        super().__init__(pipeline=pipeline, refs=refs, **params)
        self.param.watch(self.update, [p for p in self.param if p not in ('rerender', 'selection_expr', 'name')])
        self.download.view = self
        if self.selection_group:
            self._init_link_selections()
        self._initialized = False

    def __panel__(self) -> Viewable:
        if not self._initialized:
            self.update()
        return pn.panel(pn.bind(lambda e: self.panel, self.param.rerender))

    def _update_loading(self, event):
        if self._panel is not None:
            with immediate_dispatch():
                self._panel.loading = event.new

    def _update_ref(self, pname: str, ref: str, *events: param.parameterized.Event) -> None:
        # Note: Do not trigger update in View if Pipeline references
        # the same variable and is not set up to auto-update
        for event in events:
            if self.pipeline.auto_update:
                continue
            refs = []
            current = self.pipeline
            while current is not None:
                for ref in current.refs:
                    for subref in VARIABLE_RE.findall(ref):
                        if subref not in refs:
                            refs.append(subref)
                current = current.pipeline
            if any(ref == event.name for e in events for ref in refs):
                return
        super()._update_ref(pname, ref, *events)

    def _init_link_selections(self):
        doc = pn.state.curdoc or self.pipeline
        if self._ls is not None:
            return
        if doc not in View._selections and self.selection_group:
            View._selections[doc] = {}
        self._ls = View._selections.get(doc, {}).get(self.selection_group)
        if self._ls is None:
            from holoviews.selection import link_selections
            self._ls = link_selections.instance()
            if self.selection_group:
                View._selections[doc][self.selection_group] = self._ls
        if 'selection_expr' in self.param:
            self._ls.param.watch(self._update_selection_expr, 'selection_expr')

    def _update_selection_expr(self, event: param.parameterized.Event):
        self.selection_expr = event.new

    def __bool__(self) -> bool:
        return self._requires_source or self._cache is None or len(self._cache) > 0

    @catch_and_notify
    def _update_panel(self, *events: param.parameterized.Event):
        """
        Updates the cached Panel object and returns a boolean value
        indicating whether a rerender is required.
        """
        self._sync_refs(trigger=False)
        if self._panel is not None:
            self._cleanup()
            try:
                self._stream()
                return False
            except NotImplementedError:
                updates = self._get_params()
                if updates is not None:
                    self._panel.param.set_param(**updates)
                    return False
        self._panel = self.get_panel()
        return True

    @classmethod
    def _validate_filters(cls, *args, **kwargs):
        return cls._validate_list_subtypes('filters', Filter, *args, **kwargs)

    @classmethod
    def _validate_transforms(cls, *args, **kwargs):
        return cls._validate_list_subtypes('transforms', Transform, *args, **kwargs)

    @classmethod
    def _validate_sql_transforms(cls, *args, **kwargs):
        return cls._validate_list_subtypes('sql_transforms', SQLTransform, *args, **kwargs)

    @classmethod
    def _validate_pipeline(cls, *args, **kwargs):
        return cls._validate_str_or_spec('pipeline', Pipeline, *args, **kwargs)

    ##################################################################
    # Public API
    ##################################################################

    @classmethod
    def from_spec(
        cls, spec: Dict[str, Any] | str, source=None, filters=None, pipeline=None
    ) -> 'View':
        """
        Resolves a View specification given the schema of the Source
        it will be filtering on.

        Parameters
        ----------
        spec: dict
            Specification declared as a dictionary of parameter values.
        source: lumen.sources.Source
            The Source object containing the tables the View renders.
        filters: list(lumen.filters.Filter)
            A list of Filter objects which provide query values for
            the Source.
        pipeline: lumen.pipeline.Pipeline
            The Lumen pipeline driving this View. Must not be supplied
            if the spec contains a pipeline definition or reference.

        Returns
        -------
        The resolved View object.
        """
        if isinstance(spec, str):
            raise ValueError(
                "View cannot be materialized by reference. Please pass "
                "full specification for the View."
            )
        spec = spec.copy()
        resolved_spec, refs = {}, {}

        # Resolve pipeline
        if 'pipeline' in spec:
            if pipeline is not None:
                raise ValueError(
                    "Either specify the pipeline as part of the specification "
                    "or pass it in explicitly, not both."
                )
            pipeline = spec['pipeline']
            if isinstance(pipeline, str):
                resolved_spec['pipeline'] = state.pipelines[pipeline]
            else:
                resolved_spec['pipeline'] = Pipeline.from_spec(pipeline)
        else:
            overrides = {
                p: spec.pop(p) for p in Pipeline.param if p != 'name' and p in spec
            }
            for ts in ('transforms', 'sql_transforms'):
                if ts in overrides:
                    overrides[ts] = [Transform.from_spec(t) for t in overrides[ts]]
            if pipeline is None:
                if isinstance(source, str):
                    source = state.sources[source]
                pipeline = Pipeline(source=source, **overrides)
            elif 'table' in overrides and len(overrides) == 1:
                if pipeline.table != overrides['table']:
                    raise ValidationError(
                        'Table declared on view does not match table declared '
                        'on pipeline.', spec, 'table'
                    )
            elif overrides:
                pipeline = pipeline.chain(
                    filters=overrides.get('filters', []),
                    transforms=overrides.get('transforms', []),
                    sql_transforms=overrides.get('sql_transforms', []),
                    _chain_update=True
                )
            resolved_spec['pipeline'] = pipeline

        # Resolve View parameters
        view_type = View._get_type(spec.pop('type', None))
        for p, value in spec.items():
            if p in resolved_spec:
                continue
            elif p not in view_type.param:
                resolved_spec[p] = value
                continue
            parameter = view_type.param[p]
            if is_ref(value):
                refs[p] = value
                value = state.resolve_reference(value)
            if isinstance(parameter, param.ObjectSelector) and parameter.names:
                try:
                    value = parameter.names.get(value, value)
                except Exception:
                    pass
            resolved_spec[p] = value

        # Resolve download options
        download_spec = spec.pop('download', {})
        if isinstance(download_spec, str):
            *filenames, ext = download_spec.split('.')
            filename = '.'.join(filenames) or None
            download_spec = {'filename': filename, 'format': ext}
        resolved_spec['download'] = Download.from_spec(download_spec)

        view = view_type(refs=refs, **resolved_spec)

        if filters is None:
            filters = view.pipeline.traverse('filters')
            for pipeline in state.pipelines.values():
                for filt in pipeline.traverse('filters'):
                    if filt not in filters:
                        filters.append(filt)

        # Resolve ParamFilter parameters
        for filt in (filters or []):
            if isinstance(filt, ParamFilter):
                if not isinstance(filt.parameter, str):
                    continue
                name, parameter = filt.parameter.split('.')
                if name == view.name and parameter in view.param:
                    filt.parameter = view.param[parameter]
        return view

    def get_data(self):
        """
        Queries the Source for the specified table applying any
        filters and transformations specified on the View. Unlike
        `get_value` this should be used when multiple return values
        are expected.

        Returns
        -------
        DataFrame
            The queried table after filtering and transformations are
            applied.
        """
        if self._cache is not None:
            return self._cache
        if self.pipeline.data is None:
            self.pipeline._update_data()
        self._cache = data = self.pipeline.data
        if self.limit is not None:
            data = data.iloc[:self.limit]
        return data.copy()

    def get_value(self, field: str | None = None):
        """
        Queries the Source for the data associated with a particular
        field applying any filters and transformations specified on
        the View. Unlike `get_data` this method returns a single
        scalar value associated with the field and should therefore
        only be used if only a single.

        Parameters
        ----------
        field: str (optional)
            The field from the table to return; if None uses
            field defined on the View.

        Returns
        -------
        object
            A single scalar value representing the current value of
            the queried field.
        """
        data = self.get_data()
        if not len(data) or (field is not None and field not in data.columns):
            return None
        row = data.iloc[-1]
        return row[self.field if field is None else field]

    def to_spec(self, context: Dict[str, Any] | None = None) -> Dict[str, Any]:
        """
        Exports the full specification to reconstruct this component.

        Parameters
        ----------
        context: Dict[str, Any]
          Context contains the specification of all previously serialized components,
          e.g. to allow resolving of references.

        Returns
        -------
        Declarative specification of this component.
        """
        spec = super().to_spec(context)
        spec.update(self.kwargs)
        if context is None:
            return spec
        for name, pipeline in context.get('pipelines', {}).items():
            if spec.get('pipeline') == pipeline:
                spec['pipeline'] = name
                return spec
        if 'pipeline' in spec:
            if 'pipelines' not in context:
                context['pipelines'] = {}
            context['pipelines'][self.pipeline.name] = spec['pipeline']
            spec['pipeline'] = self.pipeline.name
        return spec

    def update(self, *events: param.parameterized.Event, invalidate_cache: bool=True):
        """
        Triggers an update in the View.

        Parameters
        ----------
        events: tuple
            param events that may trigger an update.
        invalidate_cache : bool
            Whether to clear the View's cache.
        """
        if invalidate_cache:
            self._cache = None
        stale = self._update_panel()
        self._initialized = True
        if stale:
            self.param.trigger('rerender')

    @property
    def control_panel(self) -> Param:
        return Param(
            self.param, parameters=self.controls, sizing_mode='stretch_width'
        )

    @property
    def panel(self) -> Viewable:
        if not self._initialized:
            self.update()
        panel = self._panel
        if isinstance(panel, PaneBase):
            if len(panel.layout) == 1 and panel._unpack:
                panel = panel.layout[0]
            else:
                panel = panel._layout
        if self.title:
            title_pane = pn.pane.HTML(
                f'<h3 align="center", style="margin-top: 0; margin-bottom: 0;">{self.title}</h3>',
                sizing_mode="stretch_width"
            )
        if self.download and self.title:
            download_pane = pn.panel(self.download)
            download_pane.sizing_mode = "fixed"
            return pn.Column(
                pn.Row(title_pane, download_pane), panel, sizing_mode=panel.sizing_mode
            )
        elif self.download:
            return pn.Column(self.download, panel, sizing_mode="stretch_both")
        elif self.title:
            return pn.Column(title_pane, panel, sizing_mode=panel.sizing_mode)
        return panel

    @property
    def refs(self) -> List[str]:
        refs = super().refs
        for c in self.controls:
            if c not in refs:
                refs.append(c)
        return refs

    ##################################################################
    # Component subclassable API
    ##################################################################

    def get_panel(self) -> Viewable:
        """
        Constructs and returns a Panel object which will represent a
        view of the queried table.

        Returns
        -------
        panel.Viewable
            A Panel Viewable object representing a current
            representation of the queried table.
        """
        return pn.panel(self.get_data())

    def _get_params(self) -> Dict[str, Any] | None:
        """
        Returns a dictionary of parameter values that will update the
        parameters on the existing _panel object inplace.  If not
        implemented each update will re-render the entire object by
        calling get_panel and replacing the existing object.
        """
        return None

    def _stream(self):
        """
        Streams new data to the existing _panel object. Will only
        be called if streaming is enabled.
        """
        raise NotImplementedError('View does not implement streaming.')

    def _cleanup(self):
        """
        Implements any cleanup that has to be performed when the View
        is updated.
        """



class Panel(View):
    """
    `Panel` views provide a way to declaratively wrap a Panel component.

    The `Panel` View is a very general purpose view that allows
    expressing arbitrary Panel objects as a specification. The Panel
    specification may be arbitrarily nested making it possible to
    specify entire layouts.  Additionally the Panel specification also
    supports references, including standard source and variable
    references and a custom `$data` reference that inserts the current
    data of the `View`.

    ```
    type: panel
      spec:
       type: panel.layout.Column
       objects:
         - type: pn.pane.Markdown
           object: '# My custom title'
         - type: pn.pane.DataFrame
           object: $data
    ```
    """

    spec = param.Dict()

    view_type: ClassVar[str] = 'panel'

    _requires_source: ClassVar[bool] = False

    def _resolve_spec(self, spec):
        if not isinstance(spec, dict) or 'type' not in spec:
            return spec
        spec = self.spec.copy()
        ptype = resolve_module_reference(spec.pop('type'), Viewable)
        params = {}
        for p, v in spec.items():
            if isinstance(v, dict) and 'type' in v:
                v = self._resolve_spec(v)
            elif isinstance(v, list):
                v = [self._resolve_spec(sv) for sv in v]
            elif is_ref(v):
                if v == '$data':
                    v = self.get_data()
                else:
                    v = state.resolve_reference(v)
            params[p] = v
        return ptype(**params)

    def get_panel(self):
        return self._resolve_spec(self.spec)


class StringView(View):
    """
    `StringView` renders the latest value of the field as a HTML string.
    """

    font_size = param.String(default='24pt', doc="""
        The font size of the rendered field value.""")

    view_type: ClassVar[str] = 'string'

    def get_panel(self) -> pn.pane.HTML:
        return pn.pane.HTML(**self._get_params())

    def _get_params(self) -> Dict[str, Any]:
        value = self.get_value()
        params = dict(self.kwargs)
        if value is None:
            params['object'] = 'No info'
        else:
            params['object'] = f'<p style="font-size: {self.font_size}">{value}</p>'
        return params


class IndicatorView(View):
    """
    `IndicatorView` renders the latest field value as a Panel `Indicator`.
    """

    indicator = param.Selector(objects=_INDICATORS, doc="""
        The name of the panel Indicator type.""")

    label = param.String(doc="""
        A custom label to use for the Indicator.""")

    view_type = 'indicator'

    def __init__(self, **params):
        if 'indicator' in params and isinstance(params['indicator'], str):
            params['indicator'] = _INDICATORS[params['indicator']]
        super().__init__(**params)
        name = params.get('label', params.get('field', ''))
        self.kwargs['name'] = name

    def get_panel(self):
        return self.indicator(**self._get_params())

    def _get_params(self) -> Dict[str, Any]:
        params = dict(self.kwargs)
        if 'data' in self.indicator.param:
            params['data'] = self.get_data()
        else:
            value = self.get_value()
            if (not isinstance(value, (type(None), str)) and np.isnan(value)):
                value = None
            params['value'] = value
        return params


class hvPlotBaseView(View):

    kind = param.String(default=None, doc="The kind of plot, e.g. 'scatter' or 'line'.")

    x = param.Selector(doc="The column to render on the x-axis.")

    y = param.Selector(doc="The column to render on the y-axis.")

    by = param.ListSelector(doc="The column(s) to facet the plot by.")

    groupby = param.ListSelector(doc="The column(s) to group by.")

    __abstract = True

    def __init__(self, **params):
        import hvplot.pandas  # type: ignore # noqa
        if 'dask' in sys.modules:
            try:
                import hvplot.dask  # type: ignore # noqa
            except Exception:
                pass
        if 'by' in params and isinstance(params['by'], str):
            params['by'] = [params['by']]
        if 'groupby' in params and isinstance(params['groupby'], str):
            params['groupby'] = [params['groupby']]
        super().__init__(**params)


class hvPlotUIView(hvPlotBaseView):
    """
    `hvPlotUIView` displays provides a component for exploring datasets interactively.
    """

    view_type = 'hvplot_ui'

    def _get_args(self):
        from hvplot.ui import hvPlotExplorer  # type: ignore
        params = {
            k: v for k, v in self.param.values().items()
            if k in hvPlotExplorer.param and v is not None and k != 'name'
        }
        return (self.get_data(),), dict(params, **self.kwargs)

    def __panel__(self):
        panel = self.get_panel()
        def ui(*events):
            panel._data = self.get_data()
            panel._plot()
            return panel
        return pn.bind(ui, self.param.rerender)

    def get_panel(self):
        from hvplot.ui import hvDataFrameExplorer
        args, kwargs = self._get_args()
        return hvDataFrameExplorer(*args, **kwargs)


class hvPlotView(hvPlotBaseView):
    """
    `hvPlotView` renders the queried data as a bokeh plot generated with hvPlot.

    hvPlot allows for a concise but powerful declaration of a plot via
    its simple API.
    """

    opts = param.Dict(default={}, doc="HoloViews options to apply on the plot.")

    streaming = param.Boolean(default=False, doc="""
        Whether to stream new data to the plot or rerender the plot.""")

    selection_expr = param.Parameter(doc="""
        A selection expression caputirng the current selection applied
        on the plot.""")

    view_type = 'hvplot'

    _field_params = ['x', 'y', 'by', 'groupby']

    _ignore_kwargs = ['tables']

    _supports_selections = True

    def __init__(self, **params):
        self._data_stream = None
        self._linked_objs = []
        super().__init__(**params)

    def get_plot(self, df):
        processed = {}
        for k, v in self.kwargs.items():
            if k in self._ignore_kwargs:
                continue
            if k.endswith('formatter') and isinstance(v, str) and '%' not in v:
                v = NumeralTickFormatter(format=v)
            processed[k] = v
        if self.streaming:
            processed['stream'] = self._data_stream
        plot = df.hvplot(
            kind=self.kind, x=self.x, y=self.y, by=self.by, groupby=self.groupby, **processed
        )
        plot = plot.opts(**self.opts) if self.opts else plot
        if self.selection_group or 'selection_expr' in self._param_watchers:
            plot = self._link_plot(plot)
        return plot

    def _link_plot(self, plot):
        self._init_link_selections()
        if self._ls is None:
            return plot
        linked_objs = list(self._ls._plot_reset_streams)
        plot = self._ls(plot)
        self._linked_objs += [
            o for o in self._ls._plot_reset_streams if o not in linked_objs
        ]
        return plot

    def _cleanup(self):
        if self._ls is None:
            return
        for obj in self._linked_objs:
            reset = self._ls._plot_reset_streams.pop(obj)
            sel_expr = self._ls._selection_expr_streams.pop(obj)
            self._ls._cross_filter_stream.input_streams.remove(sel_expr)
            sel_expr.clear()
            sel_expr.source = None
            reset.clear()
            reset.source = None
        self._linked_objs = []

    def get_panel(self):
        params = self._get_params()
        for s in ('width', 'height'):
            if f'frame_{s}' in self.kwargs:
                params[s] = self.kwargs[f'frame_{s}']
        return pn.pane.HoloViews(**params)

    def _get_params(self):
        df = self.get_data()
        if self.streaming:
            from holoviews.streams import Pipe  # type: ignore
            self._data_stream = Pipe(data=df)
        return dict(object=self.get_plot(df))

    def update(self, *events, invalidate_cache=True):
        """
        Triggers an update in the View.

        Parameters
        ----------
        events: tuple
            param events that may trigger an update.
        invalidate_cache : bool
            Whether to clear the View's cache.
        """
        # Skip events triggered by a parameter change on this View
        own_parameters = [self.param[p] for p in self.param]
        own_events = events and all(
            isinstance(e.obj, ParamFilter) and
            (e.obj.parameter in own_parameters or
            e.new is self._ls.selection_expr)
            for e in events
        )
        if own_events:
            return
        if invalidate_cache:
            self._cache = None
        if not self.streaming or self._data_stream is None:
            stale = self._update_panel()
            if stale:
                self.param.trigger('rerender')
        else:
            self._data_stream.send(self.get_data())


class Table(View):
    """
    `Table` renders data using the powerful Panel `Tabulator` component.

    See https://panel.holoviz.org/reference/widgets/Tabulator.html
    """

    page_size = param.Integer(default=20, bounds=(1, None), doc="""
        Number of rows to render per page, if pagination is enabled.""")

    view_type = 'table'

    _extension = 'tabulator'

    def get_panel(self):
        return pn.widgets.tables.Tabulator(**self._get_params())

    def _get_params(self):
        return dict(value=self.get_data(), disabled=True, page_size=self.page_size,
                    **self.kwargs)


class DownloadView(View):
    """
    `DownloadView` renders a button that allows downloading data as CSV, Excel, and parquet files.
    """

    filename = param.String(default='data', doc="""
      Filename of the downloaded file.""")

    format = param.ObjectSelector(default=None, objects=DOWNLOAD_FORMATS, doc="""
      The format to download the data in.""")

    kwargs = param.Dict(default={}, doc="""
      Keyword arguments passed to the serialization function, e.g.
      data.to_csv(file_obj, **kwargs).""")

    view_type = 'download'

    _required_keys = ["format"]

    def __bool__(self) -> bool:
        return True

    @catch_and_notify("Download failed")
    def _table_data(self) -> IO[Any]:
        io: IO[Any]
        if self.format in ('json', 'csv'):
            io = StringIO()
        else:
            io = BytesIO()
        data = self.get_data()
        if self.format == 'csv':
            data.to_csv(io, **self.kwargs)
        elif self.format == 'json':
            data.to_json(io, **self.kwargs)
        elif self.format == 'xlsx':
            data.to_excel(io, **self.kwargs)
        elif self.format == 'parquet':
            data.to_parquet(io, **self.kwargs)
        io.seek(0)
        return io

    def get_panel(self) -> pn.widgets.FileDownload:
        return pn.widgets.FileDownload(**self._get_params())

    def _get_params(self) -> Dict[str, Any]:
        filename = f'{self.filename}.{self.format}'
        return dict(filename=filename, callback=self._table_data, **self.kwargs)



class PerspectiveView(View):
    """
    `PerspectiveView` renders data into a Perspective widget.

    See https://panel.holoviz.org/reference/panes/Perspective.html for more details.
    """

    aggregates = param.Dict(None, allow_None=True, doc="""
        How to aggregate. For example {x: "distinct count"}""")

    columns = param.ListSelector(default=None, allow_None=True, doc="""
        A list of source columns to show as columns. For example ["x", "y"]""")

    computed_columns = param.ListSelector(default=None, allow_None=True, doc="""
        A list of computed columns. For example [""x"+"index""]""")

    column_pivots = param.ListSelector(None, allow_None=True, doc="""
        A list of source columns to pivot by. For example ["x", "y"]""")

    filters = param.List(default=None, allow_None=True, doc="""
        How to filter. For example [["x", "<", 3],["y", "contains", "abc"]]""")

    row_pivots = param.ListSelector(default=None, allow_None=True, doc="""
        A list of source columns to group by. For example ["x", "y"]""")

    selectable = param.Boolean(default=True, allow_None=True, doc="""
        Whether items are selectable.""")

    sort = param.List(default=None, doc="""
        How to sort. For example[["x","desc"]]""")

    plugin = param.ObjectSelector(default=_PerspectivePlugin.GRID.value, objects=_PerspectivePlugin.options(), doc="""
        The name of a plugin to display the data. For example hypergrid or d3_xy_scatter.""")

    theme = param.ObjectSelector(default='material', objects=_PERSPECTIVE_THEMES, doc="""
        The style of the PerspectiveViewer. For example material-dark""")

    view_type = 'perspective'

    _extension = 'perspective'

    _field_params = ['columns', 'computed_columns', 'column_pivots', 'row_pivots']

    def _get_params(self) -> Dict[str, Any]:
        df = self.get_data()
        param_values = dict(self.param.get_param_values())
        params = set(View.param) ^ set(PerspectiveView.param)
        kwargs = dict({p: param_values[p] for p in params}, **self.kwargs)
        return dict(object=df, toggle_config=False, **kwargs)

    def get_panel(self) -> pn.pane.Perspective:
        return pn.pane.Perspective(**self._get_params())


class AltairView(View):
    """
    `AltairView` provides a declarative way to render Altair charts.
    """

    chart = param.Dict(default={}, doc="Keyword argument for Chart.")

    x = param.Selector(doc="The column to render on the x-axis.")

    y = param.Selector(doc="The column to render on the y-axis.")

    marker = param.Selector(default='line', objects=[
        'area', 'bar', 'boxplot', 'circle', 'errorband', 'errorbar',
        'geoshape', 'image', 'line', 'point', 'rect', 'rule', 'square',
        'text', 'tick', 'trail'])

    encode = param.Dict(default={}, doc="Keyword arguments for encode.")

    mark = param.Dict(default={}, doc="Keyword arguments for mark.")

    transform = param.Dict(doc="""
        Keyword arguments for transforms, nested by the type of
        transform, e.g. {'bin': {'as_': 'binned', 'field': 'x'}}.""")

    project = param.Dict(doc="Keyword arguments for project.")

    properties = param.Dict(doc="Keyword arguments for properties.")

    view_type = 'altair'

    _extension = 'vega'

    def _transform_encoding(self, encoding: str, value: Any) -> Any:
        import altair as alt  # type: ignore
        if isinstance(value, dict):
            value = dict(value)
            for kw, val in value.items():
                if kw == 'scale':
                    if isinstance(val, list):
                        val = alt.Scale(range=val)
                    else:
                        val = alt.Scale(**val)
                if kw == 'tooltip':
                    val = [alt.Tooltip(**v) for v in val]
                value[kw] = val
            value = getattr(alt, encoding.capitalize())(**value)
        return value

    def _get_params(self) -> Dict[str, Any]:
        import altair as alt
        df = self.get_data()
        chart = alt.Chart(df, **self.chart)
        mark = getattr(chart, f'mark_{self.marker}')(**self.mark)
        x = self._transform_encoding('x', self.x)
        y = self._transform_encoding('y', self.y)
        encode = {k: self._transform_encoding(k, v) for k, v in self.encode.items()}
        encoded = mark.encode(x=x, y=y, **encode)
        if self.transform:
            for key, kwargs in self.transform.items():
                encoded = getattr(encoded, f'transform_{key}')(**kwargs)
        if self.project:
            encoded = encoded.project(**self.project)
        if self.properties:
            encoded = encoded.properties(**self.properties)
        return dict(object=encoded, **self.kwargs)

    def get_panel(self) -> pn.pane.Vega:
        return pn.pane.Vega(**self._get_params())


__all__ = [name for name, obj in locals().items() if isinstance(obj, type) and issubclass(obj, View)] + ["Download"]



import os
import pathlib

import pandas as pd
import param  # type: ignore

from lumen.transforms.base import (
    Count, Eval, Sum, Transform,
)


class CustomTransform(Transform):

    value = param.Selector()

    transform_type = 'test'


def test_resolve_module_type():
    assert Transform._get_type('lumen.transforms.base.Transform') is Transform


def test_transform_control_options(make_filesource):
    root = os.path.dirname(__file__)
    make_filesource(root)

    transform = Transform.from_spec({
        'type': 'test',
        'controls': [{
            'name': 'value',
            'options': ['A', 'B', 'C']
        }]
    })
    assert transform.param.value.objects == ['A', 'B', 'C']


def test_transform_control_options_by_reference(make_filesource):
    root = pathlib.Path(__file__).parent / '..' / 'sources'
    make_filesource(str(root))

    transform = Transform.from_spec({
        'type': 'test',
        'controls': [{
            'name': 'value',
            'options': '$original.test.C'
        }]
    })
    assert transform.param.value.objects == ['foo1', 'foo2', 'foo3', 'foo4', 'foo5']


def test_count_transform():
    df = pd._testing.makeMixedDataFrame()

    count_df = Count.apply_to(df)

    pd.testing.assert_frame_equal(count_df, df.count().to_frame().T)


def test_sum_transform():
    df = pd._testing.makeMixedDataFrame()[['A']]

    sum_df = Sum.apply_to(df)

    pd.testing.assert_frame_equal(sum_df, df.sum().to_frame().T)


def test_eval_transform():
    df = pd._testing.makeMixedDataFrame()[['A']]

    eval_df = Eval.apply_to(df, expr='B = table.A * 2')

    df2 = df.copy()
    df2['B'] = df.A * 2

    pd.testing.assert_frame_equal(eval_df, df2)



from lumen.auth import YamlAuthMapperPlugin
from lumen.dashboard import AuthSpec

auth_mapper = """
group:
  admins:
    email:
      - lumen@holoviz.org
      - panel@holoviz.org
      - param@holoviz.org
  users:
    email:
      - philipp@holoviz.org
"""


def test_yaml_auth_mapper_plugin(yaml_file):
    yaml_file.write(auth_mapper)
    yaml_file.seek(0)
    yaml_file.flush()

    mapper = YamlAuthMapperPlugin(yaml_file=yaml_file.name)

    original = {
        'group': ['admins', 'users']
    }
    transformed = mapper.transform(original)

    assert transformed == {
        'email': [
            'lumen@holoviz.org',
            'panel@holoviz.org',
            'param@holoviz.org',
            'philipp@holoviz.org'
        ]
    }

def test_auth(state_userinfo):
    auth = AuthSpec(spec={'email': 'lumen@holoviz.org'})
    auth2 = AuthSpec(spec={'email': 'panel@holoviz.org'})
    assert auth.authorized
    assert not auth2.authorized


def test_auth_case_sensitive(state_userinfo):
    auth = AuthSpec(spec={'email': 'Lumen@holoviz.org'})
    auth2 = AuthSpec(spec={'email': ' Lumen@holoviz.org  '})
    auth3 = AuthSpec(spec={'email': 'Lumen@holoviz.org'}, case_sensitive=True)
    assert auth.authorized
    assert auth2.authorized
    assert not auth3.authorized



import pytest

from lumen.filters.base import Filter
from lumen.validation import ValidationError


@pytest.mark.parametrize(
    "spec,msg",
    (
        (
            {"type": "constant", "field": "island", "value": "Torgersen"},
            None,
        ),
        (
            {"field": "island", "value": "Torgersen"},
            "Filter component specification did not declare a type",
        ),
        (
            {"type": "constant", "value": "Torgersen"},
            "The ConstantFilter component requires 'field' parameter to be defined",
        ),
        (
            {"type": "widget", "fields": "island"},
            "The WidgetFilter component requires 'field' parameter to be defined",
        ),
        (
            {'type': 'widget', 'widget': 'panel.widgets.IntSlider', 'field': 'year'},
            None
        ),
        (
            {'type': 'widget', 'widget': 'IntSlider', 'field': 'year'},
            "Filter could not resolve widget module reference 'IntSlider'."
        ),
    ),
    ids=["correct1", "missing_type", "missing_field_constant", "missing_field_widget", "valid_widget_ref", "wrong_widget_ref"],
)
def test_filter_Filter(spec, msg):
    if msg is None:
        assert Filter.validate(spec.copy()) == spec

    else:
        with pytest.raises(ValidationError, match=msg):
            Filter.validate(spec)



import pytest

from lumen.dashboard import Dashboard
from lumen.validation import ValidationError


@pytest.mark.parametrize(
    "sources,layouts,msg",
    (
        (
            {"penguins": {"type": "file", "tables": {"penguins": "url.com"}}},
            [{"title": "Table", "source": "penguins", "views": []}],
            None,
        ),
        (
            {"penguin": {"type": "file", "tables": {"penguins": "url.com"}}},
            [{"title": "Table", "source": "penguins", "views": []}],
            "Layout specified non-existent source 'penguins'",
        ),
        (
            {"penguins": {"type": "file", "tables": {"penguins": "url.com"}}},
            [{"title": "Table", "source": "penguin", "views": []}],
            "Layout specified non-existent source 'penguin'",
        ),
        (
            {"penguins": {"type": "file", "tables": {"penguins": "url.com"}}},
            [{"title": "Table", "source": "penguins", "view": []}],
            "Layout component specification contained unknown key 'view'",
        ),
        (
            {"penguins": {"tables": {"penguins": "url.com"}}},
            [{"title": "Table", "source": "penguins", "views": []}],
            "Source component specification did not declare a type",
        ),
        (
            {"penguins": {"type": "file", "tables": {"penguins": "url.com"}}},
            [{"source": "penguins", "views": []}],
            "The Layout component requires 'title' parameter to be defined",
        ),
    ),
    ids=[
        "correct",
        "missing_layout1",
        "missing_layout2",
        "missing_view",
        "missing_type",
        "missing_title",
    ],
)
def test_dashboard_Dashboard(sources, layouts, msg):
    spec = {"sources": sources, "layouts": layouts}

    if msg is None:
        assert Dashboard.validate(spec.copy()) == spec

    else:
        with pytest.raises(ValidationError, match=msg):
            Dashboard.validate(spec)


@pytest.mark.parametrize(
    "config,msg",
    (
        (
            {
                "theme": "dark",
                "layout": "column",
                "template": "bootstrap",
                "loading_color": "red",
            },
            None,
        ),
        (
            {"loading_color": 3},
            "Config component 'loading_color' value failed validation",
        ),
        (
            {"them": "dark"},
            "Config component specification contained unknown key 'them'",
        ),
        (
            {"theme": "purple"},
            "Config theme 'purple' could not be found",
        ),
        (
            {"template": "foobar"},
            "Config template 'foobar' not found",
        ),
        (
            {"template": "foobar.CustomTemplate"},
            "Config template 'foobar' module could not be imported errored with",
        ),
        (
            {"template": "lumen.CustomTemplate"},
            "Config template 'CustomTemplate' was not found in 'lumen' module",
        ),
        (
            {"template": "lumen.Dashboard"},
            "Config template 'lumen.Dashboard' is not a valid Panel template",
        ),
    ),
    ids=[
        "correct",
        "invalid_param_type",
        "misspell_theme",
        "invalid_theme",
        "invalid_template",
        "invalid_template_module",
        "invalid_template_cls_not_existing",
        "invalid_template_cls_wrong_type",
    ],
)
def test_dashboard_Config(config, msg):
    spec = {"config": config}

    if msg is None:
        assert Dashboard.validate(spec.copy()) == spec

    else:
        with pytest.raises(ValidationError, match=msg):
            Dashboard.validate(spec)


@pytest.mark.parametrize(
    "defaults,msg",
    (
        (
            {
                "sources": [{"type": "file", "dask": True}],
                "filters": [{"type": "constant", "sync_with_url": False}],
                "transforms": [{"type": "sort", "ascending": False}],
                "views": [{"type": "table", "page_size": 10}],
            },
            None,
        ),
        (
            {"source": [{"type": "file", "dask": True}]},
            "Defaults component specification contained unknown key 'source'",
        ),
        (
            {"sources": [{"dask": True}]},
            "Defaults must declare the type the defaults apply to",
        ),
        (
            {"sources": [{"type": "file", "dsk": True}]},
            "Default for FileSource 'dsk' parameter cannot be set as there is no such parameter",
        ),
        (
            {"sources": [{"type": "file", "dask": "purple"}]},
            "The default for FileSource 'dask' parameter failed validation",
        ),
    ),
    ids=[
        "correct",
        "key_wrong",
        "no_type",
        "parameter_does_not_exist",
        "parameter_value_wrong_type",
    ],
)
def test_dashboard_Defaults(defaults, msg):
    spec = {"defaults": defaults}

    if msg is None:
        assert Dashboard.validate(spec.copy()) == spec

    else:
        with pytest.raises(ValidationError, match=msg):
            Dashboard.validate(spec)


@pytest.mark.parametrize(
    "variables,msg",
    (
        (
            {
                "var1": {"type": "constant", "value": "foo"},
                "var2": {"type": "widget", "value": "$variables.var1", "kind": "TextInput"},
            },
            None,
        ),
        (
            [{"type": "constant", "value": "foo"}],
            "Dashboard component 'variables' key expected dict type but got list",
        ),
        (
            {"var1": {"value": "foo"}},
            "Variable component specification did not declare a type",
        ),
        (
            {"var1": {"type": "widget", "value": "foo", "throttled": "True", "kind": "TextInput"}},
            "Widget component 'throttled' value failed validation",
        ),
        (
            {"var1": {"type": "foo", "value": "foo"}},
            "Variable component specification declared unknown type 'foo'",
        ),
        (
            {"var1": {"type": "constant", "value": "$variables.bar"}},
            "Constant component 'value' references undeclared variable '\$variables",
        ),
    ),
    ids=[
        "correct",
        "wrong_type",
        "no_type",
        "wrong_parameter_type",
        "unknown_type",
        "unknown_ref",
    ],
)
def test_dashboard_Variables(variables, msg):
    spec = {"variables": variables}

    if msg is None:
        assert Dashboard.validate(spec.copy()) == spec

    else:
        with pytest.raises(ValidationError, match=msg):
            Dashboard.validate(spec)



"""
The Filter components supply query parameters used to filter the
tables returned by a Source.
"""
from __future__ import annotations

import types

from typing import (
    TYPE_CHECKING, Any, Callable, ClassVar, Dict, List, Tuple, Type,
)

import bokeh  # type: ignore
import pandas as pd
import panel as pn
import param  # type: ignore

from packaging.version import Version
from panel.util import classproperty

from ..base import MultiTypeComponent
from ..schema import JSONSchema
from ..state import state
from ..util import resolve_module_reference
from ..validation import ValidationError

if TYPE_CHECKING:
    from panel.viewable import Viewable
    from panel.widgets import Widget


class Filter(MultiTypeComponent):
    """
    `Filter` components supply the filter values used by `Source` components to query data. .
    """

    field = param.String(doc="The field being filtered.")

    label = param.String(doc="A label for the Filter.")

    schema = param.Dict(doc="""
        The JSON schema provided by the Source declaring information
        about the data to be filtered.""")

    shared = param.Boolean(default=False, doc="""
        Whether the filter is shared across all layouts.""")

    sync_with_url = param.Boolean(default=True, doc="""
        Whether to sync the filter state with the URL parameters.""")

    table = param.String(default=None, doc="""
        The table being filtered. If None applies to all tables.""")

    value = param.Parameter(doc="The current filter value.")

    filter_type: ClassVar[str | None] = None

    __abstract = True

    # Specification configuration
    _internal_params: ClassVar[List[str]] = ['name', 'schema']
    _requires_field: ClassVar[bool] = True

    def __init__(self, **params):
        super().__init__(**params)
        if state.config and state.config.sync_with_url and self.sync_with_url and pn.state.location:
            pn.state.location.sync(self, {'value': self.field}, on_error=self._url_sync_error)

    @classproperty
    def _required_keys(cls) -> List[str | Tuple[str, ...]]:  # type: ignore
        return ['field'] if cls._requires_field else []

    def _url_sync_error(self, values: Dict[str, Any]):
        """
        Called when URL syncing errors.
        """

    ##################################################################
    # Public API
    ##################################################################

    @classmethod
    def from_spec( # type: ignore
        cls, spec: Dict[str, Any] | str, source_schema: Dict[str, Dict[str, Any]],
        source_filters: Dict[str, 'Filter'] | None = None
    ) -> 'Filter':
        """
        Resolves a Filter specification given the schema of the Source
        (and optionally the table) it will be filtering on.

        Parameters
        ----------
        spec: dict or str
            Specification declared as a dictionary of parameter values.
        source_schema: dict
            A dictionary containing the JSON schema of the Source to
            be filtered on.
        source_filters: dict
            A dictionary of filters associated with the Source

        Returns
        -------
        The resolved Filter object.
        """
        if isinstance(spec, str):
            if source_filters is None:
                raise ValueError(f"Filter spec {repr(spec)} could not be resolved "
                                 "because source has not declared filters.")
            elif spec not in source_filters:
                raise ValueError(f"Filter spec {repr(spec)} could not be resolved, "
                                 f"available filters on the source include {list(source_filters)}.")
            return source_filters[spec]
        spec = dict(spec)
        filter_type = Filter._get_type(spec.pop('type', None))
        if not filter_type._requires_field:
            return filter_type(**spec)
        elif not 'field' in spec:
            raise ValueError('Filter specification must declare field to filter on.')
        field = spec['field']
        if 'label' not in spec:
            spec['label'] = field.title()
        table = spec.get('table')
        schema = None
        fields = []
        for key, table_schema in source_schema.items():
            if table is not None and key != table:
                continue
            schema = table_schema.get(field)
            fields.extend(list(table_schema))
            if schema is not None:
                break
        if not schema:
            raise ValueError(
                f"Source did not declare a schema for {field!r} filter. "
                f"Available fields include: {fields!r}."
            )
        return filter_type(schema={field: schema}, **spec)

    def to_spec(self, context: Dict[str, Any] | None = None) -> Dict[str, Any]:
        spec = super().to_spec(context=context)
        if spec.get('label', '').lower() == spec.get('field'):
            del spec['label']
        return spec

    @property
    def panel(self) -> Viewable | None:
        """
        Returns
        -------
        panel.Viewable or None
            A Panel Viewable object representing the filter.
        """
        return None

    @property
    def query(self) -> Any:
        """
        Returns
        -------
        object
            The current filter query which will be used by the
            Source to filter the data.
        """

    @classmethod
    def validate(
        cls, spec: Dict[str, Any] | str, context: Dict[str, Any] | None = None
    ) -> Dict[str, Any] | str:
        if isinstance(spec, str):
            return spec
        return super().validate(spec, context)


class ConstantFilter(Filter):
    """
    The ConstantFilter allows requesting a constant value from the
    Source.
    """

    filter_type: ClassVar[str] = 'constant'

    @property
    def query(self) -> Any:
        query = self.value
        field_type = (self.schema or {}).get(self.field, {}).get('type', None)
        if field_type == 'number' and isinstance(query, list) and len(query) == 2:
            return tuple(query)
        return query

    @property
    def panel(self) -> None:
        return None


class FacetFilter(Filter):
    """
    The FacetFilter allows faceting the data along some dimension to
    allow a single View to be exploded into multiple by grouping over
    one of the indexes.
    """

    filter_type = 'facet'

    @property
    def filters(self) -> List[ConstantFilter]:
        field_schema = self.schema[self.field]
        if 'enum' in field_schema:
            values = field_schema['enum']
        elif field_schema['type'] == 'integer':
            values = list(range(field_schema['inclusiveMinimum'],
                                field_schema['inclusiveMaximum']+1))
        else:
            raise ValueError(f'Could not facet on field {field_schema!r}, '
                             'currently only enum and integer fields are '
                             'supported.')
        return [
            ConstantFilter(field=self.field, value=value, label=self.label)
            for value in values
        ]


class BaseWidgetFilter(Filter):

    default = param.Parameter(doc="""
        The default value to use on the widget.""")

    disabled = param.Boolean(default=False, doc="""
        Whether the filter should be disabled.""")

    throttled = param.Boolean(default=True, doc="""
       If the widget has a value_throttled parameter use that instead,
       ensuring that no intermediate events are generated, e.g. when
       dragging a slider.""")

    visible = param.Boolean(default=True, doc="""
        Whether the filter should be visible.""")

    __abstract = True

    def _url_sync_error(self, values):
        value = values['value']
        if value is self.widget.value:
            return
        elif isinstance(self.widget.param.value, param.Tuple):
            self.widget.value = tuple(value)
        else:
            raise ValueError(f'URL syncing failed, value {value!r} could not be applied.')

    @property
    def panel(self) -> Widget:
        widget = self.widget.clone()
        if self.throttled and 'value_throttled' in self.widget.param:
            widget.link(self.widget, value_throttled='value')
            self.widget.link(widget, value='value')
            self.widget.link(widget, visible='visible', disabled='disabled', bidirectional=True)
        else:
            self.widget.link(widget, value='value', visible='visible', disabled='disabled', bidirectional=True)
        return widget


class WidgetFilter(BaseWidgetFilter):
    """
    `WidgetFilter` generates a Widget from the table schema provided by a Source.

    By default the widget type will be inferred from the data and
    depending on whether `multi` value selection is enabled.
    """

    empty_select = param.Boolean(default=True, doc="""
        Add an option to Select widgets to indicate no filtering.""")

    max_options = param.Integer(default=500, doc="""
        Maximum number of options to render.""")

    multi = param.Boolean(default=True, doc="""
        Whether to use a single-value or multi-value selection widget,
        e.g. for a numeric value this could be a regular slider or a
        range slider.""")

    widget = param.ClassSelector(class_=pn.widgets.Widget, doc="""
        The widget instance. When declared in a specification a full
        module path to a Panel widget class can be provided to override
        the default inferred widget.""")

    filter_type: ClassVar[str] = 'widget'

    _internal_params: ClassVar[List[str]] = ['name', 'schema']

    def __init__(self, **params):
        wtype = params.pop('widget', None)
        self._inferred_widget = wtype is None
        super().__init__(**params)
        self.widget = JSONSchema(
            schema=self.schema, sizing_mode='stretch_width', multi=self.multi,
            widgets={self.field: wtype} if wtype else {}
        )._widgets[self.field]
        if isinstance(self.widget, pn.widgets.Select) and self.empty_select:
            if self.widget.options[0] != ' ':
                self.widget.options.insert(0, ' ')
            self.widget.value = ' '
        if 'options' in self.widget.param and len(self.widget.options) > self.max_options:
            options = self.widget.options
            if isinstance(options, dict):
                options = dict(list(options.items())[:self.max_options])
            else:
                options = options[:self.max_options]
            self.param.warning(
                f'Filter {self.label!r} of type {type(self).__name__} was given {len(self.widget.options)} '
                f'options but were limited by max_options={self.max_options}. Check whether a different widget '
                'type would be more sensible or raise the max_options. '
            )
            self.widget.options = options
        self.widget.name = self.label
        self.widget.visible = self.visible
        self.widget.disabled = self.disabled
        self.widget.link(self, bidirectional=True, value='value', visible='visible', disabled='disabled')
        if self.default is not None:
            self.widget.value = self.default

    @classmethod
    def _validate_widget(cls, widget: str, spec: Dict[str, Any], context: Dict[str, Any]) -> str:
        try:
            resolve_module_reference(widget, pn.widgets.Widget)
        except Exception:
            raise ValidationError(
                f'{cls.__name__} could not resolve widget module reference {widget!r}.', spec
            )
        else:
            return widget

    def to_spec(self, context: Dict[str, Any] | None = None) -> Dict[str, Any]:
        spec = super().to_spec(context=context)
        if self._inferred_widget:
            del spec['widget']
        else:
            wtype = type(spec['widget'])
            spec['widget'] = f'{wtype.__module__}.{wtype.__name__}'
        return spec

    @property
    def query(self) -> Any:
        value = self.widget.value
        if value == ' ' and self.empty_select:
            return None
        if self.field in self.schema and self.multi:
            schema = self.schema[self.field]
            if isinstance(value, tuple) and len(value) == 2:
                low, high = value
                if (low == schema.get('inclusiveMinimum') and
                    high == schema.get('inclusiveMaximum')):
                    return
            elif isinstance(value, list) and schema['type'] == 'enum' and not (set(schema['enum']) - set(value)):
                return
        if not hasattr(self.widget.param.value, 'serialize') or value is None:
            return value
        else:
            value = self.widget.param.value.serialize(value)
            if isinstance(value, list) and isinstance(self.widget.value, tuple):
                return self.widget.value
            return value


class BinFilter(BaseWidgetFilter):
    """
    `BinFilter` is a special `WidgetFilter` that allows selecting from a set of bins.

    The `bins` must be declared from a list of tuples declaring the
    lower- and upper-bound of each bin and an optional set of `labels`
    of the same length.
    """

    bins = param.List(default=[], constant=True, doc="""
        A list of bins expressed as length two tuples.""")

    empty_select = param.Boolean(default=True, doc="""
        Add an option to Select widgets to indicate no filtering.""")

    labels = param.List(default=None, constant=True, doc="""
        A list of labels for each bin (optional).""")

    multi = param.Boolean(default=True, doc="""
        Whether to use a single-value or multi-value selection widget.""")

    filter_type: ClassVar[str] = 'bins'

    def __init__(self, **params):
        super().__init__(**params)
        if self.multi:
            widget = pn.widgets.MultiSelect
        else:
            widget = pn.widgets.Select
        if self.labels:
            options = dict(zip(self.labels, [tuple(b) for b in self.bins]))
        else:
            options = {f'{l} - {u}': (l, u) for l, u in self.bins}
        options[' '] = None
        if self.default is None:
            value = [] if self.multi else None
        else:
            value = tuple(self.default)
        self.widget = widget(name=self.label, options=options, value=value)
        self.widget.link(self, value='value', visible='visible', disabled='disabled', bidirectional=True)

    @property
    def query(self) -> Any:
        return self.widget.value


class BaseDateFilter(BaseWidgetFilter):

    mode = param.Selector(default='slider', objects=['slider', 'picker'], doc="""
        Whether to use a slider or a picker.""")

    multi = param.Boolean(default=True, doc="""
        Whether to use a single-value or multi-value/range selection widget.""")

    throttled = param.Boolean(default=True, doc="""
        Whether to throttle slider value changes.""")

    _as_date: ClassVar[bool] = False

    # Mapping from mode to a Panel widget, or to a function
    # that must return a tuple of a Panel widget and a dictionary
    # of parameter values that will override the current ones.
    _widget_mode_mapping: Dict[str, Type[Widget] | Callable[[BaseWidgetFilter], Tuple[Type[Widget], Dict]]]  = {}

    __abstract = True

    def __init__(self, **params):
        super().__init__(**params)
        widget_type = self._widget_mode_mapping[self.mode]
        param_overrides = {}
        if isinstance(widget_type, types.FunctionType):
            widget_type, param_overrides = widget_type(self)
        if param_overrides:
            self.param.set_param(**param_overrides)
        self.widget = widget_type(**self._widget_kwargs(as_date=self._as_date))
        self.widget.link(self, value='value', visible='visible', disabled='disabled', bidirectional=True)

    def _widget_kwargs(self, as_date: bool) -> Dict[str, Any]:
        field_schema = self.schema.get(self.field, {})
        kwargs = {
            'name': self.label,
            'start': pd.to_datetime(field_schema.get('inclusiveMinimum', None)),
            'end': pd.to_datetime(field_schema.get('inclusiveMaximum', None)),
        }
        if self.default is not None:
            kwargs['value'] = pd.to_datetime(self.default)
        if as_date:
            for key in ['value', 'start', 'end']:
                if key in kwargs and kwargs[key] is not None:
                    kwargs[key] = kwargs[key].date()
        return kwargs

    @property
    def panel(self) -> Widget:
        widget = self.widget.clone()
        if self.throttled and self.mode == 'slider':
            self.widget.link(widget, value='value')
            widget.link(self.widget, value_throttled='value')
        else:
            self.widget.link(widget, value='value', visible='visible', disabled='disabled', bidirectional=True)
        return widget

    @property
    def query(self) -> Any:
        return self.widget.value


# DateFilter and DatetimeFilter both implement __new__ to return
# a Parameterized class that has the correct `value` Parameter type,
# which depends on the filter type (calendar date vs. datetime) and mode
# (slider vs picker).

class DateFilter(BaseDateFilter):
    """
    `DateFilter` is a `WidgetFilter` specialized to select calendar dates.

    Depending on whether `multi` selection is enabled a single date or
    range of dates can be selected with a `DatePicker` or
    `DateRangeSlider` respectively.
    """

    filter_type: ClassVar[str] = 'date'

    def __new__(cls, **params):
        if params.get('multi', cls.param.multi.default):
            return _MultiCalendarDateFilter(**params)
        else:
            return _SingleCalendarDateFilter(**params)


class _SingleCalendarDateFilter(BaseDateFilter):

    _as_date: ClassVar[bool] = True

    _widget_mode_mapping = {
        'slider': pn.widgets.DateSlider,
        'picker': pn.widgets.DatePicker,
    }

    value = param.CalendarDate()


class _MultiCalendarDateFilter(BaseDateFilter):

    _as_date: ClassVar[bool] = True

    _widget_mode_mapping = {
        'slider': pn.widgets.DateRangeSlider,
        #'picker': pn.widgets.DateRangePicker,
    }

    value = param.CalendarDateRange()


class DatetimeFilter(BaseDateFilter):
    """
    `DatetimeFilter` is a `WidgetFilter` specialized to filter by datetimes.

    Depending on whether `multi` selection is enabled a single date or
    range of dates can be selected with a `DatetimePicker` or
    `DatetimeRangeSlider` respectively.
    """

    filter_type: ClassVar[str] = 'datetime'

    def __new__(cls, **params):
        if params.get('multi', cls.param.multi.default):
            return _MultiDatetimeFilter(**params)
        else:
            return _SingleDatetimeFilter(**params)


def _fallback_to_datetimepicker(inst):
    inst.param.warning(
        'Datetime multi/range slider filter not yet available, '
        'fallback to using a picker filter.'
    )
    return pn.widgets.DatetimePicker, {'mode': 'picker'}


class _SingleDatetimeFilter(BaseDateFilter):

    _as_date: ClassVar[bool] = False

    _widget_mode_mapping = {
        'slider': _fallback_to_datetimepicker,
        'picker': pn.widgets.DatetimePicker,
    }

    value = param.Date()


def _handle_datetimerangeslider(inst):
    if Version(bokeh.__version__) <= Version('2.4.3'):
        inst.param.warning(
            'Datetime multi/range slider filter requires Bokeh >= 2.4.3, '
            'fallback to using a picker filter.'
        )
        return pn.widgets.DatetimeRangePicker, {'mode': 'picker'}
    else:
        return pn.widgets.DatetimeRangeSlider, {}


class _MultiDatetimeFilter(BaseDateFilter):

    _as_date: ClassVar[bool] = False

    _widget_mode_mapping = {
        'slider': _handle_datetimerangeslider,
        'picker': pn.widgets.DatetimeRangePicker,
    }

    value = param.DateRange()


class ParamFilter(Filter):
    """
    `ParamFilter` reflects the value of a parameter declared on a `View`.

    The `ParamFilter` can be used to implement cross-filtering between
    different views.
    """

    parameter = param.ClassSelector(default=None, class_=(param.Parameter, str), doc="""
        Reference to a Parameter on an existing View.""")

    filter_type: ClassVar[str] = 'param'

    _requires_field: ClassVar[bool] = False

    def __init__(self, **params):
        super().__init__(**params)
        self.param.watch(self._attach_watcher, 'parameter')
        self._attach_watcher()

    def _update_value(self, event: param.parameterized.Event):
        self.value = event.new

    def _attach_watcher(self, *args):
        if isinstance(self.parameter, param.Parameter):
            self.parameter.owner.param.watch(self._update_value, self.parameter.name)


__all__ = [name for name, obj in locals().items() if isinstance(obj, type) and issubclass(obj, Filter)]



from __future__ import annotations

import datetime as dt
import textwrap

from typing import Any, ClassVar

import numpy as np
import param  # type: ignore

from jinja2 import Template

from .base import Transform


class SQLTransform(Transform):
    """
    Base class for SQL transforms.
    Mainly for informational purposes.
    """

    __abstract = True

    @classmethod
    def apply_to(cls, sql_in, **kwargs):
        """
        Calls the apply method based on keyword arguments passed to define transform.

        Parameters
        ----------
        sql_in: SQL Select statement to input to transformation.

        Returns
        -------
        SQL statement after application of transformation.
        """
        return cls(**kwargs).apply(sql_in)

    def apply(self, sql_in):
        """
        Given an SQL statement, manipulate it, and return a new SQL statement.

        Parameters
        ----------
        sql_in: string
            The initial SQL query to be manipulated.

        Returns
        -------
        string
            New SQL query derived from the above query.
        """
        return sql_in

    @classmethod
    def _render_template(cls, template: str, **params: Any) -> str:
        template = textwrap.dedent(template).lstrip()
        return Template(template, trim_blocks=True, lstrip_blocks=True).render(**params)


class SQLGroupBy(SQLTransform):
    """
    Performs a Group-By and aggregation
    """

    by = param.List(doc="""
        Columns to Group by""")

    aggregates = param.Dict(doc="""
        mapping of Aggregate Functions to use to which column to use them on""")

    transform_type: ClassVar[str] = 'sql_group_by'

    def apply(self, sql_in):
        template = """
            SELECT
                {{by_cols}}, {{aggs}}
            FROM ( {{sql_in}} )
            GROUP BY {{by_cols}}"""
        by_cols = ', '.join(self.by)
        aggs = []
        for agg, cols in self.aggregates.items():
            if isinstance(cols, str):
                cols = [cols]
            for col in cols:
                aggs.append(f'{agg}({col}) AS {col}')
        return self._render_template(template, by_cols=by_cols, aggs=', '.join(aggs), sql_in=sql_in)


class SQLLimit(SQLTransform):
    """
    Performs a LIMIT SQL operation on the query
    """

    limit = param.Integer(default=1000, doc="Limit on the number of rows to return")

    transform_type: ClassVar[str] = 'sql_limit'

    def apply(self, sql_in):
        template = """
            SELECT
                *
            FROM ( {{sql_in}} )
            LIMIT {{limit}}
        """
        return self._render_template(template, sql_in=sql_in, limit=self.limit)


class SQLDistinct(SQLTransform):

    columns = param.List(default=[], doc="Columns to return distinct values for.")

    transform_type: ClassVar[str] = 'sql_distinct'

    def apply(self, sql_in):
        template = """
            SELECT DISTINCT
                {{columns}}
            FROM ( {{sql_in}} )"""
        return self._render_template(template, sql_in=sql_in, columns=', '.join(self.columns))


class SQLMinMax(SQLTransform):

    columns = param.List(default=[], doc="Columns to return min/max values for.")

    transform_type: ClassVar[str] = 'sql_minmax'

    def apply(self, sql_in):
        aggs = []
        for col in self.columns:
            aggs.append(f'MIN({col}) as {col}_min')
            aggs.append(f'MAX({col}) as {col}_max')
        template = """
            SELECT
                {{columns}}
            FROM ( {{sql_in}} )"""
        return self._render_template(template, sql_in=sql_in, columns=', '.join(aggs))


class SQLColumns(SQLTransform):

    columns = param.List(default=[], doc="Columns to return.")

    transform_type: ClassVar[str] = 'sql_columns'

    def apply(self, sql_in):
        template = """
            SELECT
                {{columns}}
            FROM ( {{sql_in}} )
        """
        return self._render_template(template, sql_in=sql_in, columns=', '.join(self.columns))


class SQLFilter(SQLTransform):
    """
    Translates Lumen Filter query into a SQL WHERE statement.
    """

    conditions = param.List(doc="""
      List of filter conditions expressed as tuples of the column
      name and the filter value.""")

    transform_type: ClassVar[str] = 'sql_filter'

    @classmethod
    def _range_filter(cls, col, v1, v2):
        start = str(v1) if isinstance(v1, dt.date) else v1
        end = str(v2) if isinstance(v2, dt.date) else v2
        if isinstance(v1, dt.date) and not isinstance(v1, dt.datetime):
            start += ' 00:00:00'
        if isinstance(v2, dt.date) and not isinstance(v2, dt.datetime):
            end += ' 23:59:59'
        return f'{col} BETWEEN {start!r} AND {end!r}'

    def apply(self, sql_in):
        conditions = []
        for col, val in self.conditions:
            if val is None:
                condition = f'{col} IS NULL'
            elif np.isscalar(val):
                condition = f'{col} = {val!r}'
            elif isinstance(val, dt.datetime):
                condition = f'{col} = {str(val)!r}'
            elif isinstance(val, dt.date):
                condition = f"{col} BETWEEN '{str(val)} 00:00:00' AND '{str(val)} 23:59:59'"
            elif (isinstance(val, list) and all(
                    isinstance(v, tuple) and len(v) == 2 for v in val
            )):
                val = [v for v in val if v is not None]
                if not val:
                    continue
                condition = ' OR '.join([
                    self._range_filter(col, v1, v2) for v1, v2 in val
                ])
            elif isinstance(val, list):
                if not val:
                    continue
                non_null = [v for v in val if v is not None]
                str_non_null = "', '".join(n.replace("'", '\\\'') for n in non_null)
                condition = f"{col} IN ('{str_non_null}')"
                if not non_null:
                    condition = f'{col} IS NULL'
                elif len(val) != len(non_null):
                    condition = f'({condition}) OR ({col} IS NULL)'
            elif isinstance(val, tuple):
                condition = self._range_filter(col, *val)
            else:
                self.param.warning(
                    'Condition {val!r} on {col!r} column not understood. '
                    'Filter query will not be applied.'
                )
                continue
            conditions.append(condition)
        if not conditions:
            return sql_in

        template = """
            SELECT
                *
            FROM ( {{sql_in}} )
            WHERE ( {{conditions}} )"""
        return self._render_template(template, sql_in=sql_in, conditions=' AND '.join(conditions))


__all__ = [name for name, obj in locals().items() if isinstance(obj, type) and issubclass(obj, SQLTransform)]



import pathlib

import panel as pn
import param  # type: ignore
import yaml

from panel import Param
from panel.widgets import PasswordInput

from ..state import state
from ..util import expand_spec
from ..variables import Variable, Variables
from .base import WizardItem


class VariablesEditor(WizardItem):
    """
    Configure variables to use in your dashboard.
    """

    disabled = param.Boolean(default=True)

    enabled = param.List()

    ready = param.Boolean(default=True)

    secure = param.Boolean(default=False)

    variable_name = param.String(doc="Enter a name for the variable")

    variable_type = param.Selector(doc="Select the type of variable")

    variables = param.Dict(default={})

    layout = param.Selector(precedence=1)

    path = param.Foldername()

    _template = """
      <span style="font-size: 2em"><b>Variable Editor</b></span>
      <p>{{ __doc__ }}</p>
      <fast-divider></fast-divider>
      <div style="display: flex;">
      <form role="form" style="flex: 20%; max-width: 250px; line-height: 2em;">
        <div style="display: grid;">
          <label for="variable-name-${id}"><b>{{ param.variable_name.label }}</b></label>
          <fast-text-field id="variable-name" placeholder="{{ param.variable_name.doc }}" value="${variable_name}">
          </fast-text-field>
        </div>
        <div style="display: grid;">
          <label for="secure-${id}"><b>{{ param.secure.label }}</b></label>
          <fast-switch id="secure" checked="${secure}"></fast-switch>
        </div>
        <div style="display: flex;">
          <div style="display: grid; flex: auto;">
            <label for="type-${id}">
              <b>{{ param.variable_type.label }}</b>
            </label>
            <fast-select id="variable-select" style="min-width: 150px;" value="${variable_type}">
              {% for stype in param.variable_type.objects %}
              <fast-option id="variable-option-{{ loop.index0 }}" value="{{ stype }}">{{ stype.title() }}</fast-option>
              {% endfor %}
            </fast-select>
            <fast-tooltip anchor="type-${id}">{{ param.variable_type.doc }}</fast-tooltip>
          </div>
          <fast-button id="submit" appearance="accent" style="margin-top: auto; margin-left: 1em; width: 20px;" onclick="${_add_variable}" disabled="${disabled}">
            <b style="font-size: 2em;">+</b>
          </fast-button>
        </div>
      </form>
      <div id="variables" style="display: flex; flex-wrap: wrap; flex: 75%; margin-left: 1em;">
        {% for variable in variables.values() %}
        <div id="variable-container" style="margin: 0.5em; position: relative;">
        <fast-switch id="selected-{{ variable.name }}" {% if variable.name in enabled %}checked{% endif %} style="position: absolute; right: 0px; z-index: 100;" onchange="${script('toggle')}"></fast-switch>
        ${variable}
        </div>
        {% endfor %}
      </div>
    </div>
    """ # noqa

    _dom_events = {'variable-name': ['keyup']}

    _scripts = {
        'toggle': """
          var variable = event.target.id.split('-')[1]
          var enabled = [...data.enabled]
          if (enabled.indexOf(variable) > -1) {
            if (!event.target.checked) {
              enabled.splice(enabled.indexOf(variable), 1)
            }
          } else if (event.target.checked) {
            enabled.push(variable)
          }
          data.enabled = enabled
        """
    }

    _glob_pattern = '*.y*ml'

    def __init__(self, **params):
        super().__init__(**params)
        variables = param.concrete_descendents(Variable)
        self.param.variable_type.objects = types = [
            variable.variable_type or f'{variable.__module__}.{variable.__name__}'
            for variable in variables.values()
        ]
        if self.variable_type is None and types:
            self.variable_type = types[0]
        self._variables = state._variables[pn.state.curdoc] = Variables()
        path = params.get('path', self.path)
        components = pathlib.Path(path).glob(self._glob_pattern)
        for source in components:
            with open(source, encoding='utf-8') as f:
                yaml_spec = f.read()
                spec = yaml.safe_load(expand_spec(yaml_spec))
            if not spec:
                continue
            for varname, varspec in spec.items():
                self._add_from_spec(dict(name=varname, **varspec))

    def _get_varspec(self, var):
        variable = self._variables._vars[var]
        if variable.materialize:
            variable = variable.as_materialized()
        varspec = {
            k: v for k, v in variable.param.values().items()
            if k != 'value' and v != variable.param[k].default
        }
        vartype = type(variable)
        if variable.variable_type and vartype.__module__.startswith('lumen.'):
            varspec['type'] = variable.variable_type
        else:
            varspec['type'] = f'{vartype.__module__}.{vartype.__name__}'
        return varspec

    @param.depends('enabled', watch=True)
    def _update_enabled(self, *events):
        self.spec.clear()
        var_specs = {var: self._get_varspec(var) for var in self.enabled}
        self.spec.update(var_specs)
        ready = True
        for name in self.enabled:
            var = self._variables._vars[name]
            if var.required and not var.value:
                ready = False
        self.ready = ready

    @param.depends('variable_name', watch=True)
    def _enable_add(self):
        self.disabled = not bool(self.variable_name)

    def _add_variable(self, event=None):
        spec = {
            'type': self.variable_type,
            'name': self.variable_name,
            'secure': self.secure
        }
        self._add_from_spec(spec)
        self.variable_name = ''

    def _add_from_spec(self, spec, enable=False):
        varname, vartype, secure = spec['name'], spec['type'], spec.get('secure', False)
        enabled = spec.pop('enabled', enable)
        variable = Variable.from_spec(spec, self._variables)
        self._variables.add_variable(variable)
        if 'key' in variable.param and not variable.key:
            variable.key = variable.name
        variable.default = variable.value
        params = [
            p for p in variable.param if p not in (
                'name', 'materialize', 'required', 'secure', 'value'
            )
        ]
        widgets = {}
        if secure:
            widgets['default'] = PasswordInput
        var_editor = Param(variable, parameters=params, widgets=widgets).layout
        var_editor[0].value = f'{varname} ({vartype})'
        self.variables[varname] = var_editor
        variable.param.watch(self._update_enabled, list(variable.param))
        if enabled:
            self.enabled = self.enabled + [varname]
        self.param.trigger('variables')



from __future__ import annotations

import sys

from functools import partial
from itertools import product
from typing import (
    Any, ClassVar, Dict, List, Tuple, Type,
)

import numpy as np
import panel as pn
import param  # type: ignore
import tqdm  # type: ignore

from panel.viewable import Viewer
from panel.widgets import Widget
from typing_extensions import Literal

from .base import Component
from .filters.base import Filter, ParamFilter, WidgetFilter
from .sources.base import Source
from .state import state
from .transforms.base import Filter as FilterTransform, Transform
from .transforms.sql import SQLTransform
from .util import VARIABLE_RE, catch_and_notify, get_dataframe_schema
from .validation import ValidationError, match_suggestion_message


def auto_filters(schema: Dict[str, Dict[str, Any]]) -> List[Dict[str, str]]:
    """
    Automatically generates filter specifications from a schema.

    Arguments
    ---------
    schema:
      A schema describing the types of various fields.

    Returns
    -------
    filter_specs: A list of filter specifications.
    """
    filters = {}
    for field, fschema in schema.items():
        ftype = fschema['type']
        if ftype in ('number', 'integer'):
            fmin = fschema['inclusiveMinimum']
            fmax = fschema['inclusiveMaximum']
            if (fmin == fmax or not np.isfinite(fmin) or not np.isfinite(fmax)):
                continue
        elif ftype == 'enum' and len(fschema['enum']):
            continue
        filters[field] = {'type': 'widget', 'field': field}
    return filters


class DataFrame(param.DataFrame):
    """
    DataFrame parameter that resolves data on access.
    """

    def __get__(self, obj, objtype):
        if obj is not None and (obj.__dict__.get(self._internal_name) is None or (obj._stale and obj.auto_update)):
            obj._update_data(force=True)
        return super().__get__(obj, objtype)


def expand_queries(values, groups=('filters', 'variables')):
    spec_groups = []
    for group in groups:
        if group in values:
            var_names, var_values = zip(*values[group].items())
            variable_space = (dict(zip(var_names, vs)) for vs in product(*var_values))
        else:
            variable_space = [{}]
        spec_groups.append(list(variable_space))
    return [dict(zip(groups, group)) for group in product(*spec_groups)]


class Pipeline(Viewer, Component):
    """
    `Pipeline` encapsulates filters and transformations applied to a
    :class:`lumen.sources.base.Source` table.

    A `Pipeline` ingests data from a
    :class:`lumen.sources.base.Source` table or another `Pipeline`
    applying the declared :class:`lumen.filters.base.Filter`,
    :class:`lumen.transforms.base.Transform` and
    :class:`lumen.transforms.sql.SQLTransform` definitions. It can be
    used to drive one or more visual outputs or leveraged as a
    standalone component to encapsulate multiple data processing
    steps.
    """

    data = DataFrame(doc="The current data on this source.")

    schema = param.Dict(doc="The schema of the input data.")

    auto_update = param.Boolean(default=True, constant=True, doc="""
        Whether changes in filters, transforms and references automatically
        trigger updates in the data or whether an update has to be triggered
        manually using the update event or the update button in the UI."""
    )

    source = param.ClassSelector(class_=Source, constant=True, doc="""
        The Source this pipeline is fed by."""
    )

    pipeline = param.ClassSelector(class_=None, doc="""
        Optionally a pipeline may be chained to another pipeline."""
    )

    filters = param.List(item_type=Filter, doc="""
        A list of Filters to apply to the source data."""
    )

    sql_transforms = param.List(item_type=SQLTransform, doc="""
        A list of SQLTransforms to apply to the source data."""
    )

    transforms = param.List(item_type=Transform, doc="""
        A list of Transforms to apply to the source data."""
    )

    table = param.String(doc="""
        The name of the table driving this pipeline."""
    )

    update = param.Event(label='Apply update', doc="""
        Update event trigger (if manual update is set)."""
    )

    _stale = param.Boolean(default=False, precedence=-1, doc="""
        Whether the pipeline is stale."""
    )

    _internal_params: ClassVar[List[str]] = ['data', 'name', 'schema', '_stale']
    _required_fields: ClassVar[List[str | Tuple[str, str]]] = [('source', 'pipeline')]

    def __init__(self, *, source, table, **params):
        if 'schema' not in params:
            params['schema'] = source.get_schema(table)
        if any(isinstance(t, SQLTransform) for t in params.get('transforms', [])):
            raise TypeError('Pipeline.transforms must be regular Transform components, not SQLTransform.')
        super().__init__(source=source, table=table, **params)
        self._update_widget = pn.Param(self.param['update'], widgets={'update': {'button_type': 'success'}})[0]
        self._init_callbacks()

    def _update_stale(self, event):
        if event.new:
            self._stale = event.new

    @param.depends('_stale', watch=True)
    def _handle_stale(self):
        self._update_widget.button_type = 'warning' if self._stale else 'success'

    def _init_callbacks(self):
        self.param.watch(self._update_data, ['filters', 'sql_transforms', 'transforms', 'table', 'update'])
        self.source.param.watch(self._update_data, self.source._reload_params)
        for filt in self.filters:
            filt.param.watch(self._update_data, ['value'])
        for transform in self.transforms+self.sql_transforms:
            transform.param.watch(self._update_data, list(transform.param))
            for fp in transform._field_params:
                if isinstance(transform.param[fp], param.Selector):
                    transform.param[fp].objects = list(self.schema)
        if self.pipeline is not None:
            self.pipeline.param.watch(self._update_stale, '_stale')
            self.pipeline.param.watch(self._update_data, 'data')

    def _update_refs(self, *events: param.parameterized.Event):
        self._update_data()

    def __panel__(self) -> pn.Row:
        return pn.Row(self.control_panel, self.param.data)

    def to_spec(self, context: Dict[str, Any] | None = None) -> Dict[str, Any]:
        """
        Exports the full specification to reconstruct this component.

        Parameters
        ----------
        context: Dict[str, Any]
          Context contains the specification of all previously
          serialized components, e.g. to allow resolving of
          references.

        Returns
        -------
        Declarative specification of this component.
        """
        spec = super().to_spec(context=context)
        if 'pipeline' in spec and 'source' in spec:
            del spec['source']
        if context is None:
            return spec
        if 'pipelines' not in context:
            context['pipelines'] = {}
        for type_name in ('pipeline', 'source'):
            if type_name not in spec:
                continue
            obj = spec.pop(type_name)
            plural = f'{type_name}s'
            obj_name = getattr(self, type_name).name
            if plural not in context:
                context[plural] = {}
            context[plural][obj_name] = obj
            spec[type_name] = obj_name
            break
        context['pipelines'][self.name] = spec
        return spec

    @property
    def refs(self) -> List[str]:
        refs = self.source.refs.copy()
        for filt in self.filters:
            for ref in filt.refs:
                if ref not in refs:
                    refs.append(ref)
        for transform in self.sql_transforms+self.transforms:
            for ref in transform.refs:
                if ref not in refs:
                    refs.append(ref)
        return refs

    def _compute_data(self):
        query = {}

        # Compute Filter query
        for filt in self.filters:
            filt_query = filt.query
            if (filt_query is not None and not getattr(filt, 'disabled', None) and
                (filt.table is None or filt.table == self.table)):
                query[filt.field] = filt_query

        if self.pipeline is None:
            # Compute SQL transform expression
            if self.sql_transforms:
                if not self.source._supports_sql:
                    raise ValidationError(
                        'Can only use sql transforms source that support them. '
                        f'Found source typed {self.source.source_type!r} instead.'
                    )
                query['sql_transforms'] = self.sql_transforms
            data = self.source.get(self.table, **query)
        else:
            pipelines = []
            current = self
            while current.pipeline is not None:
                pipelines.append(current.pipeline)
                current = current.pipeline
            if self.pipeline.data is None or any(p._stale for p in pipelines):
                self.pipeline._update_data(force=True)
            data = FilterTransform.apply_to(
                self.pipeline.data, conditions=list(query.items())
            )

        # Apply ParamFilter
        for filt in self.filters:
            if not isinstance(filt, ParamFilter):
                continue
            from holoviews import Dataset  # type: ignore
            if filt.value is not None:
                ds = Dataset(data)
                data = ds.select(filt.value).data

        # Apply transforms
        for transform in self.transforms:
            data = transform.apply(data)
        return data

    @catch_and_notify
    def _update_data(self, *events: param.parameterized.Event, force: bool = False):
        if self._update_widget.loading:
            return
        if not force and not self.auto_update and not self.update:
            self._stale = True
            return

        self._update_widget.loading = True

        # Ensure all refs are up-to-date before we run the pipeline
        for f in self.filters+self.transforms+self.sql_transforms:
            f._sync_refs()

        try:
            if pn.state.curdoc:
                hold = pn.state.curdoc.callbacks.hold_value
                pn.state.curdoc.hold()
            self.data = self._compute_data()
            if state.config and state.config.on_update:
                pn.state.execute(partial(state.config.on_update, self))
        except Exception as e:
            raise e
        else:
            self._stale = False
        finally:
            self._update_widget.loading = False
            if pn.state.curdoc and not hold:
                pn.state.curdoc.unhold()

    @classmethod
    def _validate_source(
        cls, source_spec: Dict[str, Any] | str, spec: Dict[str, Any], context: Dict[str, Any]
    ) -> Dict[str, Any] | str:
        if isinstance(source_spec, str):
            if source_spec not in context['sources']:
                msg = f'Pipeline specified non-existent source {source_spec!r}.'
                msg = match_suggestion_message(source_spec, list(context['sources']), msg)
                raise ValidationError(msg, spec, source_spec)
            return source_spec
        return Source.validate(source_spec, context)

    @classmethod
    def _validate_pipeline(
        cls, pipeline_spec: Dict[str, Any] | str, spec: Dict[str, Any], context: Dict[str, Any]
    ) -> Dict[str, Any] | str:
        if isinstance(pipeline_spec, str):
            if pipeline_spec not in context['pipelines']:
                msg = f'Pipeline specified non-existent pipeline {pipeline_spec!r}.'
                msg = match_suggestion_message(pipeline_spec, list(context['pipelines']), msg)
                raise ValidationError(msg, spec, pipeline_spec)
            return pipeline_spec
        return Pipeline.validate(pipeline_spec, context)

    @classmethod
    def _validate_filters(
        cls,
        filter_specs: Dict[str, Dict[str, Any] | str] | List[Dict[str, Any] | str] | Literal['auto'],
        spec: Dict[str, Any],
        context: Dict[str, Any]
    ) -> Dict[str, Dict[str, Any] | str] | List[Dict[str, Any] | str]:
        if filter_specs == 'auto':
            return filter_specs
        for filter_spec in (filter_specs if isinstance(filter_specs, list) else filter_specs.values()):
            if not isinstance(filter_spec, str):
                continue
            elif not isinstance(spec['source'], str):
                raise ValidationError(
                    'Pipeline may only reference filters by name if the Source has been '
                    'defined by reference. Please simply move the filter definition from '
                    'the Source to the Pipeline.', spec, filter_spec
                )
            source = context['sources'][spec['source']]
            if 'filters' not in source:
                raise ValidationError(
                    f'Pipeline could not resolve {filter_spec!r} filter on {spec["source"]} source, '
                    'the source does not define any filters. ', spec, filter_spec
                )
            elif filter_spec not in source['filters']:
                msg = f'Pipeline could not resolve {filter_spec!r} filter on {spec["source"]} source.'
                msg = match_suggestion_message(filter_spec, list(source['filters']), msg)
                raise ValidationError(msg, spec, filter_spec)
        return cls._validate_dict_or_list_subtypes('filters', Filter, filter_specs, spec, context)

    @classmethod
    def validate(
        cls, spec: Dict[str, Any] | str, context: Dict[str, Any] | None = None
    ) -> Dict[str, Any] | str:
        if isinstance(spec, str):
            if context is not None and spec not in context['pipelines']:
                msg = f'Referenced non-existent pipeline {spec!r}.'
                msg = match_suggestion_message(spec, list(context['pipelines']), msg)
                raise ValidationError(msg, spec, spec)
            return spec
        return super().validate(spec, context)

    @classmethod
    def from_spec(
        cls, spec: Dict[str, Any] | str, source: Source | None = None,
        source_filters: Dict[str, Filter] | None = None
    ):
        """
        Creates a Pipeline from a specification.

        Parameters
        ----------
        spec : dict or str
            Specification declared as a dictionary of parameter values
            or a string referencing a source in the sources dictionary.

        Returns
        -------
        Resolved and instantiated Pipeline object
        """
        if isinstance(spec, str):
            return state.pipelines[spec]
        elif isinstance(spec, Pipeline):
            return spec

        spec = spec.copy()
        if source is not None:
            spec['source'] = source
        params = dict(spec)

        # Resolve source
        if 'source' in spec:
            source = spec['source']
            if isinstance(source, dict):
                resolved_source = Source.from_spec(source)
            elif isinstance(source, str):
                if source in state.sources:
                    resolved_source = state.sources[source]
                else:
                    resolved_source = state.load_source(source, state.spec['sources'][source])
            else:
                resolved_source = source  # type: ignore
            params['source'] = resolved_source
        elif 'pipeline' in spec:
            pipeline = spec['pipeline']
            if isinstance(pipeline, dict):
                pipeline = Pipeline.from_spec(pipeline)
            elif isinstance(pipeline, str) and pipeline in state.pipelines:
                pipeline = state.pipelines[pipeline]
            else:
                raise ValidationError('Pipeline {pipeline!r} could not be resolved.', spec)
            params['pipeline'] = pipeline
            params['source'] = resolved_source = pipeline.source
            params['table'] = pipeline.table

        # Validate table
        table = params.get('table')
        tables = resolved_source.get_tables()
        if table is None:
            if len(tables) > 1:
                raise ValidationError(
                    "The Pipeline specification does not contain a table and the "
                    "supplied Source has multiple tables. Please specify one of the "
                    f"following tables: {tables} in the pipeline specification.",
                    spec
                )
            params['table'] = table = tables[0]
        elif table not in tables:
            if hasattr(resolved_source, '_get_source'):
                # Certain sources perform fuzzy matching so we use the
                # internal API to see if the table has a match
                resolved_source._get_source(table)
            else:
                raise ValidationError(
                    "The Pipeline specification references a table that is not "
                    "available on the specified source. ", spec
                )

        # Resolve filters
        params['filters'] = filters = []
        filter_specs = spec.pop('filters', {})
        if filter_specs:
            params['schema'] = schema = resolved_source.get_schema(table)
        if filter_specs == 'auto':
            filter_specs = auto_filters(schema)
        for filt_spec in (filter_specs.items() if isinstance(filter_specs, dict) else filter_specs):
            if isinstance(filt_spec, tuple):
                filt_spec = dict(filt_spec[1], table=table, name=filt_spec[0])
            else:
                filt_spec = dict(filt_spec, table=table)
            filt = Filter.from_spec(filt_spec, {table: schema}, source_filters)
            filters.append(filt)

        # Resolve transforms
        transform_specs = spec.pop('transforms', [])
        params['transforms'] = [Transform.from_spec(tspec) for tspec in transform_specs]
        sql_transform_specs = spec.pop('sql_transforms', [])
        params['sql_transforms'] = [Transform.from_spec(tspec) for tspec in sql_transform_specs]

        # Instantiate and populate pre-cache
        precache = params.pop('precache', None)
        obj = cls(**params)
        if precache:
            obj.precache(precache)
        return obj

    def add_filter(
        self, filt: Filter | Type[Filter] | Widget, field: str | None = None, **kwargs
    ):
        """
        Add a filter to the pipeline.

        Arguments
        ---------
        filt: Filter | Type[Filter]
           The filter instance or filter type to add.
        field: str | None
           The field to filter on (required to instantiate Filter type).
        """
        if isinstance(filt, str):
            filt = Filter._get_type(filt)
        elif isinstance(filt, Widget):
            filt = WidgetFilter(
                widget=filt, field=field, table=self.table,
                schema={self.table: self.schema}
            )
        if not isinstance(filt, Filter):
            tspec = f'{filt.__module__}.{filt.__name__}'
            filt = Filter.from_spec(
                dict({'type': tspec, 'field': field, 'table': self.table}, **kwargs),
                {self.table: self.schema}
            )
        self.filters.append(filt)
        filt.param.watch(self._update_data, ['value'])
        self._stale = True

    def add_transform(self, transform: Transform, **kwargs):
        """
        Add a (SQL)Transform to the pipeline.

        Arguments
        ---------
        filt: Transform
           The Transform instance to add.
        """
        if isinstance(transform, str):
            transform = Transform._get_type(transform)(**kwargs)
        if isinstance(transform, SQLTransform):
            self.sql_transforms.append(transform)
        else:
            self.transforms.append(transform)
        fields = list(self.schema)
        for fparam in transform._field_params:
            transform.param[fparam].objects = fields
            transform.param.update(**{fparam: kwargs.get(fparam, fields)})
        transform.param.watch(self._update_data, list(transform.param))
        self._stale = True

    def chain(
        self,
        filters: List[Filter] | None = None,
        transforms: List[Transform] | None = None,
        sql_transforms: List[Transform] | None = None,
        **kwargs
    ):
        """
        Chains additional filtering, transform and sql_transform operations
        on an existing pipeline. Note that if one or more sql_transforms
        are provided the pipeline is cloned rather than applying the
        operations on top of the existing pipeline.

        Arguments
        ---------
        filters: List[Filter] | None
          Additional filters to apply on top of existing pipeline.
        transforms: List[Transform] | None
          Additional transforms to apply on top of existing pipeline.
        sql_transforms: List[SQLTransform] | None
          Additional filters to apply on top of existing pipeline.

        Returns
        -------
        Pipeline
        """
        if sql_transforms:
            params = {
                'filters': self.filters + (filters or []),
                'transforms': self.transforms + (transforms or []),
                'sql_transforms': self.sql_transforms + (sql_transforms or []),
                'data': None
            }
        else:
            params = {
                'filters': filters or [],
                'transforms': transforms or [],
                'sql_transforms': [],
                'pipeline': self,
                'schema': get_dataframe_schema(self.data)['items']['properties'],
                'data': None
            }
        chain_update = kwargs.pop('_chain_update', False)
        params.update(kwargs)
        new = self.clone(**params)
        if chain_update and not new.auto_update:
            self.param.watch(lambda e: new.param.trigger('update'), 'update')
        return new

    def clone(self, **params) -> 'Pipeline':
        """
        Create a new instance of the pipeline with optionally overridden parameter values.
        """
        return type(self)(**dict({p: v for p, v in self.param.values().items()
                                  if p != 'name'}, **params))

    def precache(
        self, queries: Dict[str, Dict[str, List[Any]]] | List[Dict[str, Dict[str, Any]]]
    ) -> None:
        """
        Populates the cache of the :class:`lumen.sources.base.Source` with the provided queries.

        Queries can be provided in two formats:

          - A dictionary containing 'filters' and 'variables'
            dictionaries each containing lists of values to compute
            a cross-product for, e.g.

              {
                'filters': {
                  <filter>': ['a', 'b', 'c', ...],
                  ...
                },
                'variables': {
                  <variable>: [0, 2, 4, ...],
                  ...
                }
              }
          - A list containing dictionaries of explicit values
            for each filter and variables.

              [{
                 'filters': {<filter>: 'a'},
                 'variables': {<variable>: 0}
               },
               {
                 'filters': {<filter>: 'a'},
                 'variables': {<variable>: 1}
               },
               ...
              ]
        """
        if not self.source.cache_dir or not self.precache:
            return

        if not isinstance(queries, list):
            queries = expand_queries(queries)

        old_auto = self.auto_update
        self.auto_update = False
        restore = None
        for qspec in tqdm.tqdm(queries, leave=True, file=sys.stdout):
            try:
                previous = self._set_spec(qspec)
                if restore is None:
                    restore = previous
                self._update_data(force=True)
            except Exception as e:
                raise e
        if restore:
            self._set_spec(restore)
            self._update_data(force=True)
        self.auto_update = old_auto

    def _set_spec(self, spec: Dict[str, Any]):
        previous: Dict[str, Dict[str, Any]] = {'variables': {}, 'filters': {}}
        filters = self.traverse('filters')
        for var_name, var_val in spec.get('variables', {}).items():
            variable = state.variables._vars[var_name]
            previous['variables'][var_name] = variable.value
            variable.value = var_val
        filt_spec = spec.get('filters', {})
        for filt in filters:
            if filt.field in filt_spec:
                previous['filters'][filt.field] = filt.value
                filt.value = filt_spec[filt.field]
        return previous

    def traverse(self, type) -> List[Transform] | List[Filter]:
        """
        Returns all Filter or Transform objects in a potentially chained
        pipeline.
        """
        if type not in ('filters', 'transforms', 'sql_transforms'):
            raise TypeError(f'May only traverse Pipeline filters, transforms or sql_transforms, not {type}')
        objects = []
        pipeline = self
        while pipeline is not None:
            objects.extend(getattr(pipeline, type))
            pipeline = pipeline.pipeline
        return objects

    @property
    def control_panel(self) -> pn.Column:
        col = pn.Column()
        filters = [filt.panel for filt in self.traverse('filters')]
        if any(filters):
            col.append('<div style="font-size: 1.5em; font-weight: bold;">Filters</div>')
        col.extend([filt for filt in filters if filt is not None])
        transforms = self.traverse('transforms')+self.traverse('sql_transforms')
        controls = [t.control_panel for t in transforms if t.controls]
        if controls:
            col.append('<div style="font-size: 1.5em; font-weight: bold;">Transforms</div>')
        col.extend(controls)
        variables, variable_controls = [], []
        for transform in transforms:
            for ref in transform.refs:
                var_refs = VARIABLE_RE.findall(ref)
                for var_ref in var_refs:
                    if var_ref not in variables:
                        variables.append(var_ref)
        for var_ref in variables:
            vpanel = state.variables._vars[var_ref].panel
            if vpanel is not None:
                variable_controls.append(vpanel)
        if variable_controls:
            col.append('<div style="font-size: 1.5em; font-weight: bold;">Variables</div>')
        col.extend(variable_controls)
        if not self.auto_update:
            col.append(self._update_widget)
        return col


Pipeline.param.pipeline.class_ = Pipeline



"""
Implements the baseclasses for all Component types in Lumen.

The base classes implement the core validation logic and the ability
to dynamically resolve Source and Variable references.
"""
from __future__ import annotations

import warnings

from functools import partial
from typing import (
    Any, ClassVar, Dict, List, Tuple, Type,
)

import pandas as pd
import panel as pn
import param  # type: ignore

from panel.io.cache import _container_hash, _hash_funcs
from panel.util import classproperty
from typing_extensions import Literal

from .state import state
from .util import (
    VARIABLE_RE, cleanup_expr, is_ref, resolve_module_reference,
)
from .validation import (
    ValidationError, match_suggestion_message, reverse_match_suggestion,
    validate_parameters,
)


class Component(param.Parameterized):
    """
    Baseclass for all Lumen component types including Source, Filter,
    Transform, Variable and View types. Components must implement
    serialization and deserialization into a specification dictionary
    via the `from_spec` and `to_spec` protocol. Additonally they
    should implement validation.
    """

    __abstract = True

    # Whether the component allows references
    _allows_refs: ClassVar[bool] = True

    # Parameters that are computed internally and are not part of the
    # component specification
    _internal_params: ClassVar[List[str]] = ['name']

    # Keys that must be declared declared as a list of strings or
    # tuples of strings if one of multiple must be defined.
    _required_keys: ClassVar[List[str | Tuple[str, ...]]] = []

    # Keys that are valid to define
    _valid_keys: ClassVar[List[str] | Literal['params'] | None] = None

    # Whether to enforce parameter validation in specification
    _validate_params: ClassVar[bool] = False

    def __init__(self, **params):
        self._refs = params.pop('refs', {})
        expected = list(self.param)
        validate_parameters(params, expected, type(self).name)
        if self._allows_refs:
            params = self._extract_refs(params, self._refs)
        super().__init__(**params)
        for p, ref in self._refs.items():
            if isinstance(state.variables, dict):
                continue
            elif isinstance(ref, str) and '$variables' in ref:
                ref_vars = VARIABLE_RE.findall(ref)
                state.variables.param.watch(partial(self._update_ref, p, ref), ref_vars)
                if '.' not in p and p not in params:
                    self._update_ref(p, ref)

    def _extract_refs(self, params: Dict[str, Any], refs: Dict[str, Any]):
        from .variables import Variable
        processed = {}
        for pname, pval in params.items():
            if is_ref(pval):
                refs[pname] = pval
            elif isinstance(pval, (pn.widgets.Widget, param.Parameter, Variable)):
                var = state.variables.add_variable(pval)
                processed[pname] = var.value
                refs[pname] = f'$variables.{var.name}'
                continue
            elif isinstance(pval, dict):
                subrefs = {}
                processed[pname] = self._extract_refs(pval, subrefs)
                for subkey, subref in subrefs.items():
                    refs[f'{pname}.{subkey}'] = subref
                continue
            else:
                processed[pname] = pval
        return processed

    def _update_ref(self, pname: str, ref: str, *events: param.parameterized.Event):
        """
        Component should implement appropriate downstream events
        following a change in a variable.
        """
        expr = cleanup_expr(ref)
        new_value = pd.eval(expr, local_dict=dict(state.variables), engine='python')
        if '.' in pname:
            pname, *keys = pname.split('.')
            old = getattr(self, pname)
            current = new = old.copy()
            for k in keys[:-1]:
                current = current[k]
            current[keys[-1]] = new_value
        else:
            new = new_value
        self.param.update({pname: new})

    def _sync_refs(self, trigger: bool = True):
        updates = []
        for p, ref in self._refs.items():
            if isinstance(state.variables, dict):
                continue
            elif isinstance(ref, str) and '$variables' in ref:
                with param.discard_events(self):
                    self._update_ref(p, ref)
                    pname, *_ = p.split('.')
                    updates.append(pname)
        if trigger:
            self.param.trigger(*updates)

    ##################################################################
    # Validation API
    ##################################################################

    @classproperty
    def _valid_keys_(cls) -> List[str] | None:
        if cls._valid_keys == 'params':
            return [p for p in cls.param if p not in cls._internal_params]
        else:
            return cls._valid_keys

    @classmethod
    def _validate_keys_(cls, spec: Dict[str, Any]):
        valid_keys = cls._valid_keys_
        for key in spec:
            if valid_keys is None or key in valid_keys:
                continue
            msg = f'{cls.__name__} component specification contained unknown key {key!r}.'
            msg = match_suggestion_message(key, cls._valid_keys_ or [], msg)
            raise ValidationError(msg, spec, key)

    @classmethod
    def _validate_required_(
        cls, spec: Dict[str, Any], required: List[str | Tuple[str, ...]] | None = None
    ):
        if required is None:
            required_keys = cls._required_keys
        else:
            required_keys = required
        for key in required_keys:
            if isinstance(key, str):
                if key in spec:
                    continue
                msg = f'The {cls.__name__} component requires {key!r} parameter to be defined.'
                msg, attr = reverse_match_suggestion(key, list(spec), msg)
                raise ValidationError(msg, spec, attr)
            elif isinstance(key, tuple):
                if any(f in spec for f in key):
                    continue
                skey = sorted(key)
                key_str = "', '".join(skey[:-1]) + f"' or '{skey[-1]}"
                msg = f"{cls.__name__} component requires one of '{key_str}' to be defined."
                for f in key:
                    msg, attr = reverse_match_suggestion(f, list(spec), msg)
                    if attr:
                        break
                raise ValidationError(msg, spec, attr)

    @classmethod
    def _validate_list_subtypes(
        cls, key: str, subtype: Type[Component], subtype_specs: List[Dict[str, Any] | str],
        spec: Dict[str, Any], context: Dict[str, Any], subcontext: List[Dict[str, Any] | str] | None = None
    ):
        if not isinstance(subtype_specs, list):
            raise ValidationError(
                f'{cls.__name__} component {key!r} key expected list type but got {type(subtype_specs).__name__}. '
                "This could be because of a missing dash in the yaml file.",
                spec, key
            )
        subtypes = []
        for subtype_spec in subtype_specs:
            subtype_spec = subtype.validate(subtype_spec, context)
            subtypes.append(subtype_spec)
            if subcontext is not None:
                subcontext.append(subtype_spec)
        return subtypes

    @classmethod
    def _validate_dict_subtypes(
        cls, key: str, subtype: Type[Component], subtype_specs: Dict[str, Dict[str, Any] | str],
        spec: Dict[str, Any], context: Dict[str, Any], subcontext: Dict[str, Any] | None = None
    ):
        if not isinstance(subtype_specs, dict):
            raise ValidationError(
                f'{cls.__name__} component {key!r} key expected dict type but got {type(subtype_specs).__name__}.',
                spec, key
            )
        subtypes = {}
        for subtype_name, subtype_spec in subtype_specs.items():
            subtypes[subtype_name] = subtype.validate(subtype_spec, context)
            if subcontext is not None:
                subcontext[subtype_name] = subtypes[subtype_name]
        return subtypes

    @classmethod
    def _validate_str_or_spec(
        cls, key: str, subtype: Type[Component], subtype_spec: Dict[str, Any] | str,
        spec: Dict[str, Any], context: Dict[str, Any]
    ):
        if isinstance(subtype_spec, str):
            if subtype_spec not in context[f'{key}s']:
                msg = f'{cls.__name__} component specified non-existent {key} {subtype_spec!r}.'
                msg = match_suggestion_message(subtype_spec, list(context[key]), msg)
                raise ValidationError(msg, spec, subtype_spec)
            return subtype_spec
        return subtype.validate(subtype_spec, context)

    @classmethod
    def _validate_dict_or_list_subtypes(
        cls, key: str, subtype: Type[Component], subtype_specs: Dict[str, Dict[str, Any] | str] | List[Dict[str, Any] | str],
        spec: Dict[str, Any], context: Dict[str, Any], subcontext: Dict[str, Any] | List[Dict[str, Any] | str] | None = None
    ):
        if isinstance(subtype_specs, list):
            assert subcontext is None or isinstance(subcontext, list)
            return cls._validate_list_subtypes(key, subtype, subtype_specs, spec, context, subcontext)
        else:
            assert subcontext is None or isinstance(subcontext, dict)
            return cls._validate_dict_subtypes(key, subtype, subtype_specs, spec, context, subcontext)

    @classmethod
    def _deprecation(
        cls, msg: str, key: str, spec: Dict[str, Any], update: Dict[str, Any]
    ):
        warnings.warn(msg, DeprecationWarning)
        if key not in spec:
            spec[key] = {}
        spec[key].update(update)

    @classmethod
    def _validate_ref(
        cls, key: str, value: Any, spec: Dict[str, Any], context: Dict[str, Any]
    ):
        refs = value[1:].split('.')
        if refs[0] == 'variables':
            if refs[1] not in context.get('variables', {}):
                msg = f'{cls.__name__} component {key!r} references undeclared variable {value!r}.'
                msg = match_suggestion_message(refs[1], list(context.get('variables', {})), msg)
                raise ValidationError(msg, spec, refs[1])
        elif refs[0] not in context.get('sources', {}):
            msg = f'{cls.__name__} component {key!r} references undeclared source {value!r}.'
            msg = match_suggestion_message(refs[1], list(context.get('sources', {})), msg)
            raise ValidationError(msg, spec, refs[1])

    @classmethod
    def _validate_param(cls, key: str, value: Any, spec: Dict[str, Any]):
        pobj = cls.param[key]
        try:
            if isinstance(pobj, param.Selector) and pobj.names and value in pobj.names:
                return
            pobj._validate(value)
        except Exception as e:
            msg = f"{cls.__name__} component {key!r} value failed validation: {str(e)}"
            raise ValidationError(msg, spec, key)

    @classmethod
    def _is_component_key(cls, key: str) -> bool:
        if key not in cls.param:
            return False
        pobj = cls.param[key]
        return (
            isinstance(pobj, param.ClassSelector) and
            isinstance(pobj.class_, type) and
            issubclass(pobj.class_, Component)
        )

    @classmethod
    def _is_list_component_key(cls, key: str) -> bool:
        if key not in cls.param:
            return False
        pobj = cls.param[key]
        return (
            isinstance(pobj, param.List) and
            isinstance(pobj.item_type, type) and
            issubclass(pobj.item_type, Component)
        )

    @classmethod
    def _validate_spec_(
        cls, spec: Dict[str, Any], context: Dict[str, Any] | None = None
    ) -> Dict[str, Any]:
        validated: Dict[str, Any] = {}
        if context is None:
            context = validated

        for key in (cls._valid_keys_ or list(spec)):
            if key not in spec:
                continue
            val = spec[key]
            if is_ref(val) and not cls._allows_refs:
                raise ValidationError(
                    f'{cls.__name__} component does not allow references but {key} '
                    f'value ({val!r}) is a reference.', spec, val
                )
            if hasattr(cls, f'_validate_{key}'):
                val = getattr(cls, f'_validate_{key}')(val, spec, context)
            elif cls._is_component_key(key):
                val = cls.param[key].class_.validate(val, context)
            elif cls._is_list_component_key(key):
                val = cls._validate_list_subtypes(
                    key, cls.param[key].item_type, val, spec, context
                )
            elif is_ref(val):
                cls._validate_ref(key, val, spec, context)
            elif key in cls.param:
                cls._validate_param(key, val, spec)
            validated[key] = val
        return validated

    ##################################################################
    # Public API
    ##################################################################

    @property
    def refs(self) -> List[str]:
        return [v for k, v in self._refs.items() if v.startswith('$variables.')]

    @classmethod
    def from_spec(cls, spec: Dict[str, Any] | str) -> 'Component':
        """
        Creates a Component instance from a specification.

        Parameters
        ----------
        spec : dict or str
            Specification declared as a dictionary of parameter values
            or a string referencing a source in the sources dictionary.

        Returns
        -------
        Resolved and instantiated Component object
        """
        if isinstance(spec, str):
            raise ValueError(
                "Component cannot be materialized by reference. Please pass "
                "full specification for the component."
            )
        return cls(**spec)

    def to_spec(self, context: Dict[str, Any] | None = None) -> Dict[str, Any]:
        """
        Exports the full specification to reconstruct this component.

        Parameters
        ----------
        context: Dict[str, Any]
          Context contains the specification of all previously serialized components,
          e.g. to allow resolving of references.

        Returns
        -------
        Declarative specification of this component.
        """
        spec = {}
        for p, value in self.param.values().items():
            if p in self._internal_params or value == self.param[p].default:
                continue
            elif self._is_component_key(p):
                pspec = value.to_spec(context=context)
                if not pspec:
                    continue
                value = pspec
            elif self._is_list_component_key(p):
                value = [
                    None if v is None else v.to_spec(context=context)
                    for v in value
                ]
            spec[p] = value
        if context is not None:
            spec.update(self._refs)
        return spec

    @classmethod
    def validate(
        cls, spec: Dict[str, Any] | str, context: Dict[str, Any] | None = None
    ) -> Dict[str, Any] | str:
        """
        Validates the component specification given the validation context.

        Arguments
        -----------
        spec: dict | str
          The specification for the component being validated (or a referene to the component)
        context: dict
          Validation context contains the specification of all previously validated components,
          e.g. to allow resolving of references.

        Returns
        --------
        Validated specification.
        """
        if isinstance(spec, str):
            return spec
        context = {} if context is None else context
        cls._validate_keys_(spec)
        cls._validate_required_(spec)
        return cls._validate_spec_(spec, context)


class MultiTypeComponent(Component):
    """
    MultiComponentType is the baseclass for extensible Lumen components.

    A `MultiTypeComponent` can be resolved using the `_get_type` method
    either by the name declared in the `<component>_type` attribute,
    where the name of the component represent the immediate subclasses
    of MultiTypeComponent. For example `class View(MultiTypeComponent)`
    should define `view_type` for all its descendants to override.

    Just as any other Component, `MultiTypeComponent` implements
    methods to construct an instance from a specification, export the
    specification of a component and the ability to validate a
    component.
    """

    __abstract = True

    _required_keys: ClassVar[List[str | Tuple[str, ...]]] = ['type']

    @classproperty
    def _valid_keys_(cls) -> List[str | Tuple[str, ...]] | None:
        if cls._valid_keys is None:
            valid = None
        elif cls._valid_keys == 'params':
            valid = list(cls.param)
        elif 'params' in cls._valid_keys:
            valid = cls._valid_keys.copy()
            valid.extend(list(cls.param))
        else:
            valid = cls._valid_keys.copy()

        if valid is not None and 'type' not in valid:
            valid.append('type')
        return valid

    @classproperty
    def _base_type(cls):
        if cls is MultiTypeComponent:
            return
        return cls.__mro__[cls.__mro__.index(MultiTypeComponent)-1]

    @classproperty
    def _component_type(cls) -> str:
        component_type = getattr(cls, f'{cls._base_type.__name__.lower()}_type')
        if component_type is not None:
            return component_type
        return f'{cls.__module__}.{cls.__name__}'

    @classmethod
    def _import_module(cls, component_type: str):
        base_type = cls._base_type
        try:
            import_name = f'lumen.{base_type.__name__.lower()}s.{component_type}'
            __import__(import_name)
        except ImportError as e:
            if e.name != import_name:
                msg = (
                    f"In order to use the {base_type.__name__.lower()} "
                    f"component '{component_type}', the '{e.name}' package "
                    "must be installed."
                )
                raise ImportError(msg)

    @classmethod
    def _get_type(
        cls, component_type: str, spec: Dict[str, Any] | None = None
    ) -> Type['MultiTypeComponent']:
        base_type = cls._base_type
        if component_type is None:
            raise ValidationError(
                f"No 'type' was provided during instantiation of {base_type.__name__} component.",
                spec
            )
        if '.' in component_type:
            return resolve_module_reference(component_type, base_type)

        cls._import_module(component_type)

        subcls_types = set()
        for subcls in param.concrete_descendents(cls).values():
            subcls_type = subcls._component_type
            if subcls_type is None:
                continue
            subcls_types.add(subcls_type)
            if subcls_type == component_type:
                return subcls

        msg = f"{base_type.__name__} component specification declared unknown type '{component_type}'."
        msg = match_suggestion_message(component_type, list(subcls_types), msg)
        raise ValidationError(msg, spec, component_type)

    ##################################################################
    # Public API
    ##################################################################

    @classmethod
    def from_spec(cls, spec: Dict[str, Any] | str) -> 'MultiTypeComponent':
        if isinstance(spec, str):
            raise ValueError(
                "MultiTypeComponent cannot be materialized by reference. Please pass "
                "full specification for the MultiTypeComponent."
            )
        spec = dict(spec)
        component_cls = cls._get_type(spec.pop('type'), spec)
        return component_cls(**spec)

    def to_spec(self, context: Dict[str, Any] | None = None) -> Dict[str, Any]:
        """
        Exports the full specification to reconstruct this component.

        Returns
        -------
        Resolved and instantiated Component object
        """
        spec = super().to_spec(context=context)
        spec['type'] = self._component_type
        return spec

    @classmethod
    def validate(
        cls, spec: Dict[str, Any] | str, context: Dict[str, Any] | None = None
    ) -> Dict[str, Any] | str:
        """
        Validates the component specification given the validation context and the path.

        Arguments
        -----------
        spec: dict | str
          The specification for the component being validated or a reference to the component.
        context: dict
          Validation context contains the specification of all previously validated components,
          e.g. to allow resolving of references.

        Returns
        --------
        Validated specification.
        """
        if isinstance(spec, str):
            return spec
        context = {} if context is None else context
        if 'type' not in spec:
            msg = f'{cls.__name__} component specification did not declare a type.'
            msg, attr = reverse_match_suggestion('type', list(spec), msg)
            raise ValidationError(msg, spec, attr)
        component_cls = cls._get_type(spec['type'], spec)
        component_cls._validate_keys_(spec)
        component_cls._validate_required_(spec)
        return component_cls._validate_spec_(spec, context)


_hash_funcs[Component] = lambda component: _container_hash(component.to_spec())



import pathlib

import panel as pn
import param  # type: ignore
import yaml

from .base import WizardItem
from .gallery import Gallery, GalleryItem
from .state import state


class DashboardGalleryItem(GalleryItem):

    thumbnail = param.Filename(default=pathlib.Path(__file__).parent / 'assets' / 'dashboard.png', precedence=-1)

    _template = """
    <div id="dashboard-item" onclick="${_launch}">
      <span style="font-size: 1.25em; font-weight: bold;">{{ name }}</span>
      <div id="details" style="margin-top: 1em;">
        ${view}
      </div>
      <p style="height: 4em;">${description}</p>
    </div>
    """

    def __init__(self, **params):
        super().__init__(**params)
        self.view = pn.pane.PNG(
            str(self.thumbnail), height=150, margin=0, align='center'
        )

    def _launch(self, event):
        self.selected = True


class DashboardGallery(WizardItem, Gallery):
    "Welcome to the Lumen Dashboard Builder!"

    auto_advance = param.Boolean(default=True)

    path = param.Foldername()

    spec = param.Dict(default={}, precedence=-1)

    _template = """
    <span style="font-size: 2em;">{{ __doc__ }}</span>
    <fast-divider style="margin: 1em 0;"></fast-divider>
    <span style="font-size: 1.5em">Select an existing dashboard, start from scratch or upload a dashboard to edit.</span>
    <fast-button id="upload" onclick="${_open_modal}" appearance="outline" style="float: right;">
        <b style="font-size: 1.5em;">+</b>
    </fast-button>
    <div id="items" style="margin: 1em 0; display: flex; flex-wrap: wrap; gap: 1em;">
    {% for item in items.values() %}
      <fast-card id="dashboard-container" class="gallery-item" style="height: 290px; width: 350px; padding: 1em;">
        ${item}
      </fast-card>
    {% endfor %}
      <fast-card id="dashboard-container-new" class="gallery-item" style="height: 290px; width: 350px; padding: 1em;">
        <div style="display: grid;">
          <span style="font-size: 1.25em; font-weight: bold;">Create new dashboard</span>
          <i id="add-button" onclick="${_create_new}" class="fa fa-plus" style="font-size: 12em; margin: 0.2em auto;" aria-hidden="true"></i>
        </div>
      </fast-card>
    </div>
    """

    _gallery_item = DashboardGalleryItem

    def __init__(self, **params):
        super().__init__(**params)
        self.editor = pn.widgets.Ace(
            language='yml', sizing_mode='stretch_both', theme='dracula',
            margin=(0, 20), height=400
        )
        self.upload = pn.widgets.FileInput(
            sizing_mode='stretch_width', margin=(0, 20), height=50
        )
        self.thumbnail_upload = pn.widgets.FileInput(
            sizing_mode='stretch_width', margin=0, name='Upload a thumbnail'
        )
        self.upload.param.watch(self._yaml_upload, 'value')
        self.namer = pn.widgets.TextInput(
            name='Dashboard name', placeholder='Give the dashboard a name',
            margin=(5, 0)
        )
        self.description = pn.widgets.TextAreaInput(
            name='Provide a description', sizing_mode='stretch_both',
            max_length=100
        )
        self.save = pn.widgets.Button(name='Save', sizing_mode='stretch_width')
        self.save.on_click(self._save)
        self._modal_content = [
            '# Upload dashboard\nUpload a yaml file',
            self.upload,
            pn.Row(
                pn.Column(self.namer, self.thumbnail_upload),
                self.description, sizing_mode='stretch_width',
                margin=(0, 20)
            ),
            self.editor,
            self.save
        ]

    def _save(self, event):
        name = self.namer.value
        spec = yaml.safe_load(self.editor.value)
        spec['metadata'] = {
            'description': self.description.value
        }
        if self.thumbnail_upload.value:
            ext = self.thumbnail_upload.filename.split('.')[-1]
            thumbnail_path = pathlib.Path(self.path) / f'{name}.{ext}'
            with open(thumbnail_path, 'wb') as f:
                f.write(self.thumbnail_upload.value)
            spec['metadata']['thumbnail'] = str(thumbnail_path)
        with open(pathlib.Path(self.path) / f'{name}.yml', 'w', encoding='utf-8') as f:
            f.write(yaml.dump(spec))
        self.items[name] = self._gallery_item(name=name, spec=spec)
        self.param.trigger('items')
        state.template.close_modal()
        self._reset_form()

    def _reset_form(self):
        self.namer.value = ''
        self.editor.value = ''
        self.description.value = ''
        self.upload.filename = ''
        self.upload.value = None
        self.thumbnail_upload.filename = ''
        self.thumbnail_upload.value = None

    def _yaml_upload(self, event):
        if event.new is None:
            return
        name = '.'.join(self.upload.filename.split('.')[:-1])
        self.namer.value = name
        self.editor.value = event.new.decode('utf-8')
        return

    def _create_new(self, event):
        self.spec = {'sources': {}, 'layouts': []}

    def _selected(self, event):
        self.spec = event.obj.spec
